"""Multi-dimensional interpolation utilitites."""

from __future__ import annotations

from collections.abc import Iterable, Sequence
from functools import partial
from itertools import product
from typing import cast

import jax
import jax.numpy as jnp
import numpy as np
import pandas as pd
import xarray as xr
from jaxtyping import Array
from natsort import natsorted
from numpy.typing import NDArray

import sax

__all__ = [
    "interpolate_xarray",
    "to_df",
    "to_xarray",
]


def interpolate_xarray(
    xarr: xr.DataArray,
    /,
    **kwargs: sax.FloatArray | str,
) -> dict[str, sax.FloatArray]:
    """Interpolate a multi-dimensional xarray with JAX.

    Args:
        xarr: the xarray to do a grid-interpolation over
        **kwargs: the other parameters to interpolate over
    """
    with jax.ensure_compile_time_eval():
        data = jnp.asarray(xarr)

        params: dict[str, NDArray] = {str(k): np.asarray(xarr[k]) for k in xarr.dims}
        new_params, strings = _extract_strings(params)
        target_name = str(xarr.dims[-1])
        targets = strings[target_name]
        if target_name in kwargs:
            msg = f"Cannot interpolate over target parameter {target_name}."
            raise ValueError(msg)
        string_kwargs = {k: v for k, v in kwargs.items() if k in strings}
        new_kwargs = {**kwargs, **{k: [str(v)] for k, v in string_kwargs.items()}}

    s, axs, pos = _evaluate_general_corner_model(
        data,
        new_params,
        strings,
        **new_kwargs,
    )

    axs_rev = dict(
        sorted([(v, k) for k, v in axs.items() if k != target_name], reverse=True)
    )

    ss = {k: s.take(pos[target_name][k], axs[target_name]) for k in targets}
    for ax, name in axs_rev.items():
        if name in string_kwargs:
            ss = {
                kk: s.take(pos[name][string_kwargs[name]], ax)  # type: ignore[reportArgumentType]
                for kk, s in ss.items()
            }

    return ss


def to_xarray(
    stacked_data: pd.DataFrame,
    *,
    target_names: Iterable[str] = ("amp", "phi"),
) -> xr.DataArray:
    """Converts a stacked dataframe into a multi-dimensional xarray.

    Args:
        stacked_data: the data to convert
        target_names: the names of the target columns.
    """
    target_names = list(target_names)
    df = stacked_data.copy()
    for name in target_names:
        if name not in df.columns:
            msg = (
                f"target name {name!r} not found in dataframe. "
                f"Got a dataframe with columns: {df.columns}."
            )
            raise ValueError(msg)
    param_names = [c for c in df.columns if c not in target_names]
    df = cast(pd.DataFrame, df[[*param_names, *target_names]])
    df = df.sort_values(by=param_names).reset_index(drop=True)
    params: dict[str, Array | NDArray] = {
        c: np.asarray(df[c].unique()) for c in param_names
    }
    params["targets"] = np.array(list(target_names), dtype=object)
    data = df[target_names].to_numpy()
    data = data.reshape(*(v.shape[0] for v in params.values()))
    data = jnp.asarray(data)
    return xr.DataArray(data=data, coords=params)


def to_df(
    obj: xr.DataArray | sax.SType,
    *,
    target_name: str = "target",
    **kwargs: sax.ArrayLike,
) -> pd.DataFrame:
    """Converts an object into a stacked dataframe.

    Args:
        obj: an xarray or a sax.SType object.
        target_name: the name of the target column
            in the dataframe (ignored when obj is an SType).
        kwargs: the coordinates of the SType values axes
            (ignored when obj is an xarray).
    """
    if isinstance(obj, xr.DataArray):
        xarr = obj
        stacked = xarr.stack(__stacked_dim=xarr.dims)  # noqa: PD013
        df = (
            stacked.reset_index("__stacked_dim")
            .to_dataframe(name=target_name)
            .reset_index(drop=True)
        )
        return df

    stype = obj
    return _sdict_to_df(stype, **kwargs)


def _evaluate_general_corner_model(
    data: Array,
    params: dict[str, Array],
    strings: dict[str, dict[str, int]],
    /,
    **kwargs: Array,
) -> tuple[Array, dict[str, int], dict[str, dict[str, int]]]:
    """Hypercube representation interpolator.

    Args:
        data: the data to interpolate (output of to_hypercube)
        params: the parameter sample points (output of to_hypercube)
        strings: non-interpolatable parameters (like strings)
            with their values (output of to_hypercube)
        **kwargs: the values at which to interpolate the s-parameter.
            These values should be broadcastable into each other.

    Returns:
        the interpolated S-matrix at the values of **kwargs
    """
    given_params = {k: kwargs.get(k, jnp.asarray(v).mean()) for k, v in params.items()}
    given_strings = {
        p: (list(v.values()) if p not in kwargs else [v[vv] for vv in kwargs[p]])
        for p, v in strings.items()
    }
    string_locations = {k: i for i, k in enumerate(given_params) if k in given_strings}
    param_locations = {
        k: i for i, k in enumerate(given_params) if k not in given_strings
    }
    data = jnp.transpose(
        data, axes=[*param_locations.values(), *string_locations.values()]
    )
    given_params = {k: v for k, v in given_params.items() if k not in given_strings}
    num_params = len(given_params)
    data = _downselect(data, list(given_strings.values()))
    param_shape, string_shape = data.shape[:num_params], data.shape[num_params:]
    data = data.reshape(*param_shape, -1)
    given_param_values = [jnp.asarray(v) for v in given_params.values()]
    stacked_params = jnp.stack(jnp.broadcast_arrays(*given_param_values), 0)
    coords = _get_coordinates(
        [v for k, v in params.items() if k in given_params], stacked_params
    )
    result = _map_coordinates(data, coords)
    axs = {k: i for i, k in enumerate(given_strings)}
    rev_strings = {k: {vv: kk for kk, vv in v.items()} for k, v in strings.items()}
    pos = {
        k: {rev_strings[k][vv]: i for i, vv in enumerate(v)}
        for k, v in given_strings.items()
    }
    return result.reshape(*string_shape, *result.shape[1:]), axs, pos


def _downselect(data: Array, idxs_list: list[list[int]]) -> Array:
    for i, idxs in enumerate(reversed(idxs_list), start=1):
        data = data.take(np.asarray(idxs, dtype=int), axis=-i)
    return data


@partial(jax.vmap, in_axes=(-1, None), out_axes=0)
def _map_coordinates(input: Array, coordinates: Sequence[Array]) -> Array:  # noqa: A002
    return jax.scipy.ndimage.map_coordinates(input, coordinates, 1, mode="nearest")


def _get_coordinate(arr1d: Array, value: Array) -> Array:
    return jnp.interp(value, arr1d, jnp.arange(arr1d.shape[0]))


def _get_coordinates(arrs1d: Sequence[Array], values: jnp.ndarray) -> list[Array]:
    # don't use vmap as arrays in arrs1d could have different shapes...
    return [_get_coordinate(a, v) for a, v in zip(arrs1d, values, strict=True)]


def _extract_strings(
    params: dict[str, Array | NDArray] | dict[str, Array] | dict[str, NDArray],
) -> tuple[dict[str, Array], dict[str, dict[str, int]]]:
    string_map = {}  # jax does not like strings
    new_params = {}
    for k, v in list(params.items()):
        if v.dtype == object:  # probably string array!
            string_map[k] = {s: i for i, s in enumerate(v)}
            new_params[k] = np.array(list(string_map[k].values()), dtype=int)
        else:
            new_params[k] = np.array(v, dtype=float)
    return new_params, string_map


def _sdict_to_df(stype: sax.SType, **coords: sax.ArrayLike) -> pd.DataFrame:
    if not coords:
        msg = "The coords of at least one dimension should be given."
        raise ValueError(msg)
    coords = {k: np.atleast_1d(v) for k, v in coords.items()}
    sdict: sax.SDict = {k: jnp.atleast_1d(v) for k, v in sax.sdict(stype).items()}
    sdict = dict(zip(sdict, jnp.broadcast_arrays(*sdict.values()), strict=True))
    shape = jnp.asarray(next(iter(sdict.values()))).shape
    if len(shape) != len(coords):
        msg = "Specify at least one array of coordinates per dimension of the sdict."
        raise ValueError(msg)
    for i, (d, (k, a)) in enumerate(zip(shape, coords.items(), strict=True)):
        if a.ndim != 1:
            msg = f"Coords should be 1D arrays. {k}.ndim = {a.ndim}"
            raise ValueError(msg)
        if a.shape[0] != d:
            msg = (
                f"the length of coord {i} [len({k})={a.shape[0]}] should match "
                f"the size of axis {i} [{d}] of each element in the sdict."
            )
            raise ValueError(msg)
    ports, modes = set(), set()
    for p1, p2 in sdict:
        for p in [p1, p2]:
            p, *m = p.split("@")
            ports.add(p)
            modes.add("".join(m))
    ports = natsorted(ports)
    modes = natsorted(modes)

    sdict = {
        k: jnp.stack([jnp.abs(v), jnp.angle(v)], axis=-1) for k, v in sdict.items()
    }
    coords["amp_phi"] = np.array(["amp", "phi"])

    dfs = []
    port_mode = lambda p, m: f"{p}@{m}" if m else p  # noqa: E731
    zero = jnp.zeros_like(next(iter(sdict.values())))
    for p1, m1, p2, m2 in product(ports, modes, ports, modes):
        pm1, pm2 = port_mode(p1, m1), port_mode(p2, m2)
        values = sdict.get((pm1, pm2), zero)
        xarr = xr.DataArray(values, coords)
        df = sax.to_df(xarr, target_name="amp")
        phi = df["amp"].to_numpy()[1::2]
        df = df.loc[::2, :].drop(columns=["amp_phi"]).reset_index(drop=True)
        df["phi"] = phi
        df["port_in"] = p1
        df["port_out"] = p2
        df["mode_in"] = m1 or "1"
        df["mode_out"] = m2 or "1"
        dfs.append(df)

    df = pd.concat(dfs, axis=0)
    columns = [*[c for c in df.columns if c not in ["amp", "phi"]], *["amp", "phi"]]

    return cast(pd.DataFrame, df[columns].reset_index(drop=True))



import sax

SAMPLE_NETLIST = {
    "instances": {
        "lft": "coupler",
        "top": "waveguide",
        "btm": "waveguide",
        "rgt": "coupler",
    },
    "connections": {
        "lft,out0": "btm,in0",
        "btm,out0": "rgt,in0",
        "lft,out1": "top,in0",
        "top,out0": "rgt,in1",
    },
    "ports": {
        "in0": "lft,in0",
        "in1": "lft,in1",
        "out0": "rgt,out0",
        "out1": "rgt,out1",
    },
}


def test_circuit() -> None:
    _mzi, _info = sax.circuit(
        netlist=SAMPLE_NETLIST,
        models={
            "coupler": sax.models.coupler_ideal,
            "waveguide": sax.models.straight,
        },
    )


def test_1port_circuit() -> None:
    """Test that 1-port circuits are supported."""
    netlist = {
        "instances": {
            "wg1": "waveguide",
        },
        "connections": {},
        "ports": {
            "in": "wg1,in0",
        },
    }

    models = {
        "waveguide": sax.models.straight,
    }

    circuit, _info = sax.circuit(netlist, models)
    result = circuit()

    # Verify the circuit has exactly 1 port
    ports = sax.get_ports(result)
    assert len(ports) == 1
    assert "in" in ports

    # Verify the S-matrix has the expected structure
    assert ("in", "in") in result


if __name__ == "__main__":
    print(test_circuit())
    print(test_1port_circuit())



"""SAX Crossing Models."""

import jax
import jax.numpy as jnp
from pydantic import validate_call

import sax


@jax.jit
@validate_call
def crossing_ideal(wl: sax.FloatArrayLike = sax.WL_C) -> sax.SDict:
    """Ideal waveguide crossing model.

    ```{svgbob}
            in1
             o2
              *
              |
              |
     o1 *-----+-----* o3
    in0       |     out0
              |
              *
             o4
           out1
    ```

    Args:
        wl: Wavelength in micrometers.

    Returns:
        The crossing s-matrix

    Examples:
        Ideal crossing:

    ```python
    # mkdocs: render
    import matplotlib.pyplot as plt
    import numpy as np
    import jax.numpy as jnp
    import sax

    sax.set_port_naming_strategy("optical")

    wl = sax.wl_c()
    s = sax.models.crossing_ideal(wl=wl)
    thru = np.abs(s[("o1", "o3")]) ** 2
    cross = np.abs(s.get(("o1", "o2"), jnp.zeros_like(wl))) ** 2
    plt.figure()
    plt.plot(wl, thru, label="thru")
    plt.plot(wl, cross, label="cross")
    plt.xlabel("Wavelength [μm]")
    plt.ylabel("Power")
    plt.legend()
    ```
    """
    one = jnp.ones_like(jnp.asarray(wl))
    p = sax.PortNamer(2, 2)
    return sax.reciprocal(
        {
            (p.o1, p.o3): one,
            (p.o2, p.o4): one,
        }
    )



"""SAX S-Matrix utilities."""

from __future__ import annotations

from functools import wraps
from typing import cast, overload

import jax
import jax.numpy as jnp
import jax.scipy as jsp
from jaxtyping import Array
from natsort import natsorted
from pydantic import validate_call

import sax

from .multimode import _consolidate_sdense

__all__ = [
    "block_diag",
    "get_mode",
    "get_modes",
    "get_port_combinations",
    "get_ports",
    "reciprocal",
    "scoo",
    "sdense",
    "sdict",
]


@overload
def sdict(S: sax.Model) -> sax.SDictModel: ...


@overload
def sdict(S: sax.SType) -> sax.SDict: ...


def sdict(S: sax.Model | sax.SType) -> sax.SDictModel | sax.SDict:
    """Convert an S-matrix to SDict (dictionary) format.

    SDict format represents S-parameters as a dictionary with port pair tuples
    as keys and complex values as entries. This is the most flexible format for
    sparse matrices and custom port naming.

    Args:
        S: S-matrix in SCoo, SDense format, or a model that returns such matrices.

    Returns:
        S-matrix in SDict format, or a model that returns SDict format.

    Raises:
        ValueError: If the input cannot be converted to SDict format.

    Example:
        ```python
        # Convert from other formats
        scoo_matrix = (Si, Sj, Sx, port_map)
        sdict_matrix = sdict(scoo_matrix)


        # Convert a model
        def my_model(wl=1.55):
            return some_sdense_matrix


        sdict_model = sdict(my_model)
        ```
    """
    if callable(_model := S):

        @wraps(_model)
        def model(**kwargs: sax.SettingsValue) -> sax.SDict:
            return sdict(_model(**kwargs))

        return model

    if isinstance(_sdict := S, dict):
        return _sdict

    if len(_scoo := cast(tuple, S)) == 4:
        return _scoo_to_sdict(*_scoo)

    if len(_sdense := cast(tuple, S)) == 2:
        return _sdense_to_sdict(*_sdense)

    msg = f"Could not convert S-matrix to sdict. Got: {S!r}."
    raise ValueError(msg)


@overload
def scoo(S: sax.Model) -> sax.SCooModel: ...


@overload
def scoo(S: sax.SType) -> sax.SCoo: ...


def scoo(S: sax.Model | sax.SType) -> sax.SCooModel | sax.SCoo:
    """Convert an S-matrix to SCoo (coordinate) format.

    SCoo format represents S-parameters in coordinate (sparse) format with
    separate arrays for row indices (Si), column indices (Sj), values (Sx),
    and port mapping. This format is memory-efficient for sparse matrices.

    Args:
        S: S-matrix in SDict, SDense format, or a model that returns such matrices.

    Returns:
        S-matrix in SCoo format (Si, Sj, Sx, port_map), or a model that returns
        SCoo format.

    Raises:
        ValueError: If the input cannot be converted to SCoo format.

    Example:
        ```python
        # Convert from dict format
        sdict_matrix = {("in", "out"): 0.9 + 0.1j, ("out", "in"): 0.9 + 0.1j}
        Si, Sj, Sx, port_map = scoo(sdict_matrix)


        # Convert a model
        def my_model(wl=1.55):
            return some_sdict_matrix


        scoo_model = scoo(my_model)
        ```
    """
    if callable(_model := S):

        @wraps(_model)
        def model(**kwargs: sax.SettingsValue) -> sax.SCoo:
            return scoo(_model(**kwargs))

        return model

    if isinstance(_sdict := S, dict):
        return _sdict_to_scoo(_sdict)

    if len(_scoo := cast(tuple, S)) == 4:
        return _scoo

    if len(_sdense := cast(tuple, S)) == 2:
        return _sdense_to_scoo(*_sdense)

    msg = f"Could not convert S-matrix to scoo. Got: {S!r}."
    raise ValueError(msg)


@overload
def sdense(S: sax.Model) -> sax.SDenseModel: ...


@overload
def sdense(S: sax.SType) -> sax.SDense: ...


def sdense(S: sax.SType | sax.Model) -> sax.SDenseModel | sax.SDense:
    """Convert an S-matrix to SDense (dense matrix) format.

    SDense format represents S-parameters as a dense complex matrix with an
    associated port mapping. This format is efficient for dense matrices and
    enables fast linear algebra operations.

    Args:
        S: S-matrix in SDict, SCoo format, or a model that returns such matrices.

    Returns:
        S-matrix in SDense format (matrix, port_map), or a model that returns
        SDense format.

    Raises:
        ValueError: If the input cannot be converted to SDense format.

    Example:
        ```python
        # Convert from dict format
        sdict_matrix = {("in", "out"): 0.9 + 0.1j, ("out", "in"): 0.9 + 0.1j}
        matrix, port_map = sdense(sdict_matrix)


        # Convert a model
        def my_model(wl=1.55):
            return some_sdict_matrix


        sdense_model = sdense(my_model)
        ```
    """
    if callable(_model := S):

        @wraps(_model)
        def model(**kwargs: sax.SettingsValue) -> sax.SDense:
            return sdense(_model(**kwargs))

        return model

    if isinstance(_sdict := S, dict):
        return _sdict_to_sdense(_sdict)

    if len(_scoo := cast(tuple, S)) == 4:
        return _scoo_to_sdense(*_scoo)

    if len(_sdense := cast(tuple, S)) == 2:
        return _sdense

    msg = f"Could not convert S-matrix to sdense. Got: {S!r}."
    raise ValueError(msg)


def reciprocal(sdict: sax.SDict) -> sax.SDict:
    """Make an SDict S-matrix reciprocal by ensuring S[i,j] = S[j,i].

    Reciprocity is a fundamental property of passive optical devices where
    the S-parameter from port i to port j equals the S-parameter from port j
    to port i. This function enforces reciprocity by copying existing values.

    Args:
        sdict: S-matrix in SDict format.

    Returns:
        Reciprocal S-matrix in SDict format with symmetric port pairs.

    Example:
        ```python
        # Make a non-reciprocal matrix reciprocal
        s_matrix = {("in", "out"): 0.9 + 0.1j}
        s_reciprocal = reciprocal(s_matrix)
        # Result: {("in", "out"): 0.9+0.1j, ("out", "in"): 0.9+0.1j}
        ```
    """
    return {
        **{(p1, p2): v for (p1, p2), v in sdict.items()},
        **{(p2, p1): v for (p1, p2), v in sdict.items()},
    }


def block_diag(*arrs: Array) -> Array:
    """Create block diagonal matrix with arbitrary batch dimensions.

    Constructs a block diagonal matrix from square input matrices. All input
    arrays must have the same batch dimensions and be square in their last
    two dimensions.

    Args:
        *arrs: Square matrices to place on the block diagonal. All must have
            matching batch dimensions.

    Returns:
        Block diagonal matrix with batch dimensions preserved.

    Raises:
        ValueError: If batch dimensions don't match or matrices aren't square.

    Example:
        ```python
        import jax.numpy as jnp

        A = jnp.array([[1, 2], [3, 4]])
        B = jnp.array([[5, 6], [7, 8]])
        C = block_diag(A, B)
        # Result: [[1, 2, 0, 0],
        #          [3, 4, 0, 0],
        #          [0, 0, 5, 6],
        #          [0, 0, 7, 8]]
        ```
    """
    batch_shape = arrs[0].shape[:-2]

    N = 0
    for arr in arrs:
        if batch_shape != arr.shape[:-2]:
            msg = "Batch dimensions for given arrays don't match."
            raise ValueError(msg)
        m, n = arr.shape[-2:]
        if m != n:
            msg = "given arrays are not square."
            raise ValueError(msg)
        N += n

    arrs = tuple(arr.reshape(-1, arr.shape[-2], arr.shape[-1]) for arr in arrs)
    batch_block_diag = jax.vmap(jsp.linalg.block_diag, in_axes=0, out_axes=0)
    block_diag = batch_block_diag(*arrs)
    return block_diag.reshape(*batch_shape, N, N)


def get_ports(S: sax.SType) -> tuple[sax.Port, ...] | tuple[sax.PortMode]:
    """Extract port names from an S-matrix.

    Returns the port names present in the S-matrix in natural sorted order.
    For multimode S-matrices, returns port@mode combinations.

    Args:
        S: S-matrix in any format (SDict, SCoo, or SDense).

    Returns:
        Tuple of port names (for single-mode) or port@mode strings (for multimode).

    Raises:
        TypeError: If input is a model function (not evaluated) or invalid type.

    Example:
        ```python
        # Single-mode matrix
        s_matrix = {("in", "out"): 0.9, ("out", "in"): 0.9}
        ports = get_ports(s_matrix)
        # Result: ("in", "out")

        # Multimode matrix
        s_mm = {("in@TE", "out@TE"): 0.9, ("in@TM", "out@TM"): 0.8}
        ports_mm = get_ports(s_mm)
        # Result: ("in@TE", "in@TM", "out@TE", "out@TM")
        ```
    """
    if callable(S):
        msg = (
            "Getting the ports of a model is no longer supported. "
            "Please Evaluate the model first: Use get_ports(model()) in stead of "
            f"get_ports(model). Got: {S}"
        )
        raise TypeError(msg)
    if isinstance(sdict := S, dict):
        ports_set = {p1 for p1, _ in sdict} | {p2 for _, p2 in sdict}
        return tuple(natsorted(ports_set))

    if isinstance(S, tuple):
        *_, pm = cast(sax.SCoo | sax.SDense, S)
        return tuple(natsorted(pm.keys()))

    msg = f"Expected an SType. Got: {S!r} [{type(S)}]"
    raise TypeError(msg)


@validate_call
def get_modes(S: sax.STypeMM) -> tuple[sax.Mode, ...]:
    """Extract the optical modes from a multimode S-matrix.

    Returns all unique optical modes (e.g., "TE", "TM") present in the
    multimode S-matrix port names.

    Args:
        S: Multimode S-matrix in any format with port@mode naming convention.

    Returns:
        Tuple of unique mode names found in the S-matrix.

    Example:
        ```python
        # Multimode S-matrix with TE and TM modes
        s_mm = {
            ("in@TE", "out@TE"): 0.9,
            ("in@TM", "out@TM"): 0.8,
            ("in@TE", "out@TM"): 0.1,
        }
        modes = get_modes(s_mm)
        # Result: ("TE", "TM")
        ```
    """
    return tuple(get_mode(pm) for pm in get_ports(S))


@validate_call
def get_mode(pm: sax.PortMode) -> sax.Mode:
    """Extract the mode from a port@mode string.

    Parses the mode identifier from the standard port@mode naming convention
    used in multimode S-matrices.

    Args:
        pm: Port-mode string in the format "port@mode".

    Returns:
        The mode identifier (part after @).

    Example:
        ```python
        mode = get_mode("waveguide1@TE")
        # Result: "TE"

        mode = get_mode("input@TM")
        # Result: "TM"
        ```
    """
    return pm.split("@")[1]


def get_port_combinations(S: sax.Model | sax.SType) -> tuple[tuple[str, str], ...]:
    """Extract all port pair combinations from an S-matrix.

    Returns all (input_port, output_port) combinations present in the S-matrix.
    This is useful for understanding the connectivity structure of the device.

    Args:
        S: S-matrix in any format (SDict, SCoo, or SDense).

    Returns:
        Tuple of (port1, port2) combinations representing S-parameter entries.

    Raises:
        TypeError: If input is a model function (not evaluated).

    Example:
        ```python
        # S-matrix with cross-coupling
        s_matrix = {
            ("in1", "out1"): 0.9,
            ("in1", "out2"): 0.1,
            ("in2", "out1"): 0.1,
            ("in2", "out2"): 0.9,
        }
        combinations = get_port_combinations(s_matrix)
        # Result:
        # (("in1", "out1"), ("in1", "out2"), ("in2", "out1"), ("in2", "out2"))
        ```
    """
    if callable(S):
        msg = (
            "Getting the port combinations of a model is no longer supported. "
            "Please Evaluate the model first: Use get_ports(model()) in stead of "
            f"get_ports(model). Got: {S}"
        )
        raise TypeError(msg)
    if isinstance(sdict := S, dict):
        return tuple(sdict.keys())
    if len(scoo := cast(sax.SCoo, S)) == 4:
        Si, Sj, _, pm = scoo
        rpm = {int(i): str(p) for p, i in pm.items()}
        return tuple(
            natsorted((rpm[int(i)], rpm[int(j)]) for i, j in zip(Si, Sj, strict=False))
        )
    if len(sdense := cast(sax.SDense, S)) == 2:
        _, pm = sdense
        return tuple(natsorted((p1, p2) for p1 in pm for p2 in pm))
    msg = "Could not extract ports for given S"
    raise ValueError(msg)


def _scoo_to_sdict(
    Si: sax.IntArray1D,
    Sj: sax.IntArray1D,
    Sx: sax.ComplexArray,
    ports_map: sax.PortMap,
) -> sax.SDict:
    sdict = {}
    inverse_ports_map = {int(i): p for p, i in ports_map.items()}
    for i, (si, sj) in enumerate(zip(Si, Sj, strict=True)):
        input_port = inverse_ports_map.get(int(si), "")
        output_port = inverse_ports_map.get(int(sj), "")
        sdict[input_port, output_port] = Sx[..., i]
    return {(p1, p2): v for (p1, p2), v in sdict.items() if p1 and p2}


def _sdense_to_sdict(S: Array, ports_map: sax.PortMap) -> sax.SDict:
    sdict = {}
    for p1, i in ports_map.items():
        for p2, j in ports_map.items():
            sdict[p1, p2] = S[..., i, j]
    return sdict


def _sdict_to_scoo(sdict: sax.SDict) -> sax.SCoo:
    all_ports = {}
    for p1, p2 in sdict:
        all_ports[p1] = None
        all_ports[p2] = None
    ports_map = {p: int(i) for i, p in enumerate(all_ports)}
    Sx = jnp.stack(jnp.broadcast_arrays(*sdict.values()), -1)
    Si = jnp.array([ports_map[p] for p, _ in sdict])
    Sj = jnp.array([ports_map[p] for _, p in sdict])
    return Si, Sj, Sx, ports_map


def _sdense_to_scoo(S: sax.ComplexArray, ports_map: sax.PortMap) -> sax.SCoo:
    S, ports_map = _consolidate_sdense((S, ports_map))
    Sj, Si = jnp.meshgrid(jnp.arange(S.shape[-1]), jnp.arange(S.shape[-2]))
    return Si.ravel(), Sj.ravel(), S.reshape(*S.shape[:-2], -1), ports_map


def _scoo_to_sdense(
    Si: sax.IntArray1D,
    Sj: sax.IntArray1D,
    Sx: sax.ComplexArray,
    ports_map: dict[str, int],
) -> sax.SDense:
    n_col = len(ports_map)
    S = jnp.zeros((*Sx.shape[:-1], n_col, n_col), dtype=complex)
    S = S.at[..., Si, Sj].add(Sx)
    return S, ports_map


def _sdict_to_sdense(sdict: sax.SDict) -> sax.SDense:
    Si, Sj, Sx, ports_map = _sdict_to_scoo(sdict)
    return _scoo_to_sdense(Si, Sj, Sx, ports_map)



"""Type conversion utilities for SAX.

This module provides utilities for safely converting between different SAX types
using Pydantic validation. It offers both strict conversion (raises errors) and
optional conversion (returns None on failure).

Examples:
    ```python
    import sax.saxtypes as sxt

    # Strict conversion
    x: int = 5
    y: sxt.Complex = sxt.into[sxt.Complex](x)
    print(y)  # (5+0j)

    # Optional conversion
    if (z := sxt.try_into[sxt.Float](x)) is not None:
        print(z)  # 5.0
    ```
"""

from __future__ import annotations

__all__ = ["into", "try_into"]

from collections.abc import Callable
from typing import Any, TypeVar, cast, get_args, overload

from pydantic import PlainValidator, TypeAdapter, ValidationError
from typing_extensions import _AnnotatedAlias

T = TypeVar("T")


class Into(type):
    """Metaclass for the `into` type converter.

    This metaclass enables the `into[Type]` syntax for type conversion.
    It creates type-safe converters that use Pydantic validation.
    """

    @overload
    def __getitem__(cls, key: type[T]) -> Callable[..., T]: ...

    @overload
    def __getitem__(cls, key: str) -> Callable[..., Any]: ...

    @overload
    def __getitem__(cls, key: Any) -> Callable[..., Any]: ...

    def __getitem__(
        cls,
        key: type[T] | str | Any,
    ) -> Callable[..., T | Any]:
        """Get a type converter for the specified key.

        Args:
            key: The target type, string name, or type specification.

        Returns:
            A converter function that validates and converts objects.
        """
        if isinstance(key, str) and hasattr(cls, key):
            return getattr(cls, key)

        if isinstance(key, _AnnotatedAlias):
            validator = get_args(key)[-1]
            if isinstance(validator, PlainValidator):
                return getattr(validator.func, "__wrapped__", validator.func)
        return _wrapped_validator(cast(type[T], key))


def _wrapped_validator(key: type[T]) -> Callable[..., T]:
    """Create a wrapped validator function for a given type.

    Args:
        key: The target type to validate against.

    Returns:
        A validator function that converts objects to the target type.

    Raises:
        TypeError: If the object cannot be converted to the target type.
    """

    def into_type(obj: Any) -> T:
        """Convert an object to the target type.

        Args:
            obj: The object to convert.

        Returns:
            The converted object.

        Raises:
            TypeError: If conversion fails.
        """
        try:
            return TypeAdapter(key).validate_python(obj, strict=False)
        except ValidationError as e:
            raise TypeError(str(e)) from e

    return into_type


class into(metaclass=Into):  # noqa: N801
    """Type converter utility for strict type conversion.

    Use this class with bracket notation to convert objects to specific SAX types.
    Raises TypeError if conversion fails.

    Examples:
        Validate and cast objects to SAX or python types:

        ```python
        import sax.saxtypes as sxt

        # Convert to different numeric types
        x = sxt.into[sxt.Complex](5)  # (5+0j)
        y = sxt.into[sxt.Float](42)  # 42.0
        z = sxt.into[sxt.ComplexArray]([1, 2, 3])  # Array([1.+0.j, 2.+0.j, 3.+0.j])

        # Convert to SAX-specific types
        port = sxt.into[sxt.Port]("in0")  # "in0"
        name = sxt.into[sxt.Name]("component1")  # "component1"
        ```
    """


class TryInto(type):
    """Metaclass for the `try_into` optional type converter.

    This metaclass enables the `try_into[Type]` syntax for optional type conversion.
    Returns None instead of raising errors when conversion fails.
    """

    @overload
    def __getitem__(cls, key: type[T]) -> Callable[..., T | None]: ...

    @overload
    def __getitem__(cls, key: str) -> Callable[..., Any | None]: ...

    @overload
    def __getitem__(cls, key: Any) -> Callable[..., Any | None]: ...

    def __getitem__(cls, key: type[T] | str | Any) -> Callable[..., T | Any | None]:
        """Get an optional type converter for the specified key.

        Args:
            key: The target type, string name, or type specification.

        Returns:
            A converter function that returns the converted object or None on failure.
        """
        from ..utils import maybe

        return maybe(into[key])


class try_into(metaclass=TryInto):  # noqa: N801
    """Optional type converter utility that returns None on failure.

    Use this class with bracket notation to safely attempt type conversion.
    Returns None instead of raising errors when conversion fails.

    Examples:
        Validate and cast objects to SAX or python types safely:

        ```python
        import sax.saxtypes as sxt

        # Safe conversion attempts
        result = sxt.try_into[sxt.Float]("3.14")  # 3.14
        result = sxt.try_into[sxt.Float]("invalid")  # None

        # Use in conditional expressions
        if (value := sxt.try_into[sxt.Complex]("1+2j")) is not None:
            print(f"Converted: {value}")  # (1+2j)

        # Handle arrays safely
        arr = sxt.try_into[sxt.FloatArray]([[1, 2], [3, 4]])
        if arr is not None:
            print(f"Array shape: {arr.shape}")  # (2, 2)
        ```
    """



"""Forward-only backend for SAX circuit evaluation."""

from __future__ import annotations

from typing import Any

import jax.numpy as jnp
import networkx as nx

import sax

__all__ = [
    "analyze_circuit_forward",
    "analyze_instances_forward",
    "evaluate_circuit_forward",
]


def analyze_instances_forward(
    instances: sax.Instances,
    models: sax.Models,
) -> dict[sax.InstanceName, sax.SCoo]:
    """Analyze circuit instances for the forward-only backend.

    Prepares instance S-matrices for the forward-only backend by converting all
    component models to SCoo (coordinate) format. This backend is specialized
    for feed-forward circuits without feedback loops.

    Args:
        instances: Dictionary mapping instance names to instance definitions
            containing component names and settings.
        models: Dictionary mapping component names to their model functions.

    Returns:
        Dictionary mapping instance names to their S-matrices in SCoo format.

    Note:
        The forward-only backend is designed for circuits with unidirectional
        signal flow and no feedback paths. It uses a simplified approach that
        may not be accurate for circuits with reflections or loops.

    Example:
        ```python
        instances = {
            "wg1": {"component": "waveguide", "settings": {"length": 10.0}},
            "amp1": {"component": "amplifier", "settings": {"gain": 20.0}},
        }
        models = {"waveguide": waveguide_model, "amplifier": amplifier_model}
        analyzed = analyze_instances_forward(instances, models)
        ```
    """
    instances = sax.into[sax.Instances](instances)
    models = sax.into[sax.Models](models)
    model_names = set()
    for i in instances.values():
        model_names.add(i["component"])
    dummy_models = {k: sax.scoo(models[k]()) for k in model_names}
    dummy_instances = {}
    for k, i in instances.items():
        dummy_instances[k] = dummy_models[i["component"]]
    return dummy_instances


def analyze_circuit_forward(
    analyzed_instances: dict[sax.InstanceName, sax.SDict],  # noqa: ARG001
    connections: sax.Connections,
    ports: sax.Ports,
) -> Any:  # noqa: ANN401
    """Analyze circuit topology for the forward-only backend.

    Prepares the circuit connection information for the forward-only backend
    evaluation. This backend assumes unidirectional signal flow and does not
    account for reflections or bidirectional coupling.

    Args:
        analyzed_instances: Instance S-matrices from analyze_instances_forward.
            Not used in this analysis step but required for interface consistency.
        connections: Dictionary mapping instance ports to each other, defining
            internal circuit connections.
        ports: Dictionary mapping external port names to instance ports.

    Returns:
        Tuple containing connections and ports information for circuit evaluation.

    Example:
        ```python
        connections = {"wg1,out": "amp1,in", "amp1,out": "wg2,in"}
        ports = {"in": "wg1,in", "out": "wg2,out"}
        analyzed = analyze_circuit_forward(analyzed_instances, connections, ports)
        ```
    """
    return connections, ports


def evaluate_circuit_forward(
    analyzed: Any,  # noqa: ANN401
    instances: dict[str, sax.SDict],
) -> sax.SDict:
    """Evaluate circuit S-matrix using forward-only propagation.

    Computes the circuit response using a simplified forward propagation approach.
    This method assumes unidirectional signal flow and uses breadth-first search
    to propagate signals through the circuit without considering reflections.

    The algorithm:
    1. Creates a directed graph representation of the circuit
    2. For each input port, injects a unit signal
    3. Uses BFS to propagate signals through the circuit
    4. Records the signal levels at output ports

    Args:
        analyzed: Circuit analysis data from analyze_circuit_forward containing
            connections and ports information.
        instances: Dictionary mapping instance names to their evaluated S-matrices
            in SDict format.

    Returns:
        Circuit S-matrix in SDict format, typically containing only forward
        transmission terms (input to output ports).

    Warning:
        This backend is only accurate for feed-forward circuits without
        reflections or feedback paths. For circuits with bidirectional coupling
        or reflections, use the Filipsson-Gunnar or KLU backends instead.

    Example:
        ```python
        # Circuit analysis and instances (feed-forward only)
        analyzed = (connections, ports)
        instances = {
            "wg1": {("in", "out"): 0.95},  # Low-loss waveguide
            "amp1": {("in", "out"): 10.0},  # 20dB amplifier
        }
        circuit_s = evaluate_circuit_forward(analyzed, instances)
        # Result contains only forward transmission terms
        ```
    """
    connections, ports = analyzed
    edges = _graph_edges_directed(instances, connections, ports)

    graph = nx.DiGraph()
    graph.add_edges_from(edges)

    # Dictionary to store signals at each node
    circuit_sdict = {}
    for in_port in ports:
        if in_port.startswith("in"):
            node_signals = {("", in_port): 1}
            bfs_output = nx.bfs_layers(graph, ("", in_port))
            for layer in bfs_output:
                layer_signals = {}
                for node in layer:
                    if node in node_signals:
                        signal = node_signals[node]
                        for neighbor in graph.successors(node):
                            transmission = graph[node][neighbor]["transmission"]
                            if neighbor in layer_signals:
                                layer_signals[neighbor] += signal * transmission
                            else:
                                layer_signals[neighbor] = signal * transmission
                node_signals.update(layer_signals)
            sdict = {
                (in_port, p2): v
                for (p1, p2), v in node_signals.items()
                if p1 == "" and p2.startswith("out")
            }
            circuit_sdict.update(sdict)
    return circuit_sdict


def _split_port(port: str) -> tuple[str, str]:
    try:
        instance, port = port.split(",")
    except ValueError:
        (port,) = port.split(",")
        instance = ""
    return instance, port


def _graph_edges_directed(
    instances: dict[str, sax.SDict],
    connections: dict[str, str],
    ports: dict[str, str],
) -> list[tuple[tuple[str, str], tuple[str, str], dict[str, Any]]]:
    one = jnp.array([1.0], dtype=float)
    edges_dict = {}
    edges_dict.update({_split_port(k): _split_port(v) for k, v in connections.items()})
    edges_dict.update({_split_port(k): _split_port(v) for k, v in ports.items()})
    edges = []
    for n1, n2 in edges_dict.items():
        if n1[0] == "" and n1[1].startswith("out"):
            edges += [(n2, n1, {"transmission": one})]
        else:
            edges += [(n1, n2, {"transmission": one})]

    for instance, s in instances.items():
        for (p1, p2), w in sax.sdict(s).items():
            if p1.startswith("in") and p2.startswith("out"):
                edges += [
                    (
                        (instance, p1),
                        (instance, p2),
                        {"transmission": jnp.asarray(w, dtype=complex).ravel()},
                    )
                ]
    return edges



"""Test execution of notebooks in the nbs directory."""

import shutil
from collections.abc import Generator
from pathlib import Path

import pytest
from papermill.engines import papermill_engines
from papermill.execute import raise_for_execution_errors
from papermill.iorw import load_notebook_node

TEST_DIR = Path(__file__).resolve().parent
NBS_DIR = TEST_DIR / "nbs"
NBS_FAIL_DIR = TEST_DIR / "failed"

shutil.rmtree(NBS_FAIL_DIR, ignore_errors=True)
NBS_FAIL_DIR.mkdir(exist_ok=True)


def _find_notebooks(*dir_parts: str) -> Generator[Path, None, None]:
    base_dir = TEST_DIR.joinpath(*dir_parts).resolve()
    for path in base_dir.rglob("*.ipynb"):
        if "checkpoint" in path.name:
            continue
        yield path


@pytest.mark.parametrize("path", sorted(_find_notebooks("nbs")))
def test_nbs(path: Path | str) -> None:
    fn = Path(path).name
    nb = load_notebook_node(str(path))
    nb = papermill_engines.execute_notebook_with_engine(
        engine_name=None,
        nb=nb,
        kernel_name="sax",
        input_path=str(path),
        output_path=None,
    )
    raise_for_execution_errors(nb, str(NBS_FAIL_DIR / fn))



"""All types and type-validators used in SAX."""

from __future__ import annotations

from .anymode import (
    Backend,
    BackendLike,
    CircuitInfo,
    Model,
    ModelFactory,
    Models,
    PortCombination,
    PortMap,
    SCoo,
    SCooModel,
    SCooModelFactory,
    SDense,
    SDenseModel,
    SDenseModelFactory,
    SDict,
    SDictModel,
    SDictModelFactory,
    SType,
)
from .core import (
    ArrayLike,
    Bool,
    BoolArray,
    BoolArrayLike,
    BoolLike,
    Complex,
    ComplexArray,
    ComplexArray1D,
    ComplexArray1DLike,
    ComplexArrayLike,
    ComplexLike,
    Float,
    FloatArray,
    FloatArray1D,
    FloatArray1DLike,
    FloatArray2D,
    FloatArray2DLike,
    FloatArrayLike,
    FloatLike,
    Int,
    IntArray,
    IntArray1D,
    IntArray1DLike,
    IntArrayLike,
    IntLike,
    IOLike,
    Name,
)
from .custom_warnings import ExperimentalWarning
from .into import (
    into,
    try_into,
)
from .multimode import (
    Mode,
    ModelFactoryMM,
    ModelMM,
    ModelsMM,
    PortCombinationMM,
    PortMapMM,
    PortMode,
    SCooMM,
    SCooModelFactoryMM,
    SCooModelMM,
    SDenseMM,
    SDenseModelFactoryMM,
    SDenseModelMM,
    SDictMM,
    SDictModelFactoryMM,
    SDictModelMM,
    STypeMM,
)
from .netlist import (
    AnyNetlist,
    Component,
    Connections,
    Instance,
    Instances,
    Net,
    Netlist,
    Nets,
    Placement,
    Placements,
    Ports,
    RecursiveNetlist,
    default_placement,
)
from .settings import (
    Settings,
    SettingsValue,
)
from .singlemode import (
    InstanceName,
    InstancePort,
    ModelFactorySM,
    ModelSM,
    ModelsSM,
    Port,
    PortCombinationSM,
    PortMapSM,
    SCooModelFactorySM,
    SCooModelSM,
    SCooSM,
    SDenseModelFactorySM,
    SDenseModelSM,
    SDenseSM,
    SDictModelFactorySM,
    SDictModelSM,
    SDictSM,
    STypeSM,
)

__all__ = [
    "AnyNetlist",
    "ArrayLike",
    "Backend",
    "BackendLike",
    "Bool",
    "BoolArray",
    "BoolArrayLike",
    "BoolLike",
    "CircuitInfo",
    "Complex",
    "ComplexArray",
    "ComplexArray1D",
    "ComplexArray1DLike",
    "ComplexArrayLike",
    "ComplexLike",
    "Component",
    "Connections",
    "ExperimentalWarning",
    "Float",
    "FloatArray",
    "FloatArray1D",
    "FloatArray1D",
    "FloatArray1DLike",
    "FloatArray1DLike",
    "FloatArray2D",
    "FloatArray2DLike",
    "FloatArrayLike",
    "FloatLike",
    "IOLike",
    "Instance",
    "InstanceName",
    "InstancePort",
    "Instances",
    "Int",
    "IntArray",
    "IntArray1D",
    "IntArray1DLike",
    "IntArrayLike",
    "IntLike",
    "Mode",
    "Model",
    "ModelFactory",
    "ModelFactoryMM",
    "ModelFactorySM",
    "ModelMM",
    "ModelSM",
    "Models",
    "ModelsMM",
    "ModelsSM",
    "Name",
    "Net",
    "Netlist",
    "Nets",
    "Placement",
    "Placements",
    "Port",
    "PortCombination",
    "PortCombinationMM",
    "PortCombinationSM",
    "PortMap",
    "PortMapMM",
    "PortMapSM",
    "PortMode",
    "Ports",
    "RecursiveNetlist",
    "SCoo",
    "SCooMM",
    "SCooModel",
    "SCooModelFactory",
    "SCooModelFactoryMM",
    "SCooModelFactorySM",
    "SCooModelMM",
    "SCooModelSM",
    "SCooSM",
    "SDense",
    "SDenseMM",
    "SDenseModel",
    "SDenseModelFactory",
    "SDenseModelFactoryMM",
    "SDenseModelFactorySM",
    "SDenseModelMM",
    "SDenseModelSM",
    "SDenseSM",
    "SDict",
    "SDictMM",
    "SDictModel",
    "SDictModelFactory",
    "SDictModelFactoryMM",
    "SDictModelFactorySM",
    "SDictModelMM",
    "SDictModelSM",
    "SDictSM",
    "SType",
    "STypeMM",
    "STypeSM",
    "Settings",
    "SettingsValue",
    "default_placement",
    "into",
    "try_into",
]



"""SAX Multimode Types and type coercions.

This module defines types and validators specifically for multi-mode optical
circuits, where ports can support multiple optical modes (e.g., TE0, TE1, TM0).
Ports are specified using the 'port@mode' notation.

References:
    Numpy type reference: https://numpy.org/doc/stable/reference/arrays.scalars.html
"""

from __future__ import annotations

__all__ = [
    "ModelFactoryMM",
    "ModelMM",
    "ModelsMM",
    "PortCombinationMM",
    "PortMapMM",
    "PortMode",
    "SCooMM",
    "SCooModelFactoryMM",
    "SCooModelMM",
    "SDenseMM",
    "SDenseModelFactoryMM",
    "SDenseModelMM",
    "SDictMM",
    "SDictModelFactoryMM",
    "SDictModelMM",
    "STypeMM",
]

from collections.abc import Callable
from typing import (
    Annotated,
    Any,
    TypeAlias,
)

from .core import ComplexArray, IntArray1D, Name, val
from .singlemode import (
    cast_string,
    val_callable_annotated,
    val_not_callable_annotated,
    val_port,
    val_sax_callable,
)


def val_mode(obj: Any) -> Mode:
    """Validate a mode specification string.

    Modes can be any string identifier, including numerical modes like '@0', '@1'
    or named modes like '@TE0', '@TM0'.

    Args:
        obj: The object to validate as a mode.

    Returns:
        The validated mode string.

    Examples:
        Validating a string as a mode:

        ```python
        import sax.saxtypes.multimode as mm

        # Valid mode specifications
        result = mm.val_mode("0")  # "0" (numerical mode)
        result = mm.val_mode("TE0")  # "TE0" (named mode)
        result = mm.val_mode("TM1")  # "TM1" (named mode)
        ```
    """
    return str(obj)  # just a string to allow '@0' etc.


Mode: TypeAlias = Annotated[str, val(val_mode)]
"""A mode identifier string (e.g., '0', 'TE0', 'TM1')."""


def val_port_mode(obj: Any) -> PortMode:
    """Validate a port-mode specification in 'port@mode' format.

    Args:
        obj: The object to validate as a port-mode specification.

    Returns:
        The validated port-mode string.

    Raises:
        TypeError: If the string doesn't follow 'port@mode' format.

    Examples:
        Validating a port-mode string:

        ```python
        import sax.saxtypes.multimode as mm

        # Valid port-mode specifications
        result = mm.val_port_mode("in0@0")  # "in0@0"
        result = mm.val_port_mode("out1@TE0")  # "out1@TE0"
        result = mm.val_port_mode("port@TM1")  # "port@TM1"
        ```
    """
    s = cast_string(obj)
    parts = s.split("@")
    if len(parts) != 2:
        msg = f"a PortMode should have exactly one '@'-separator. Got: {obj!r}"
        raise TypeError(msg)
    port, mode = parts
    port = val_port(port)
    mode = val_mode(mode)
    return f"{port}@{mode}"


PortMode: TypeAlias = Annotated[str, val(val_port_mode)]
"""A port-mode specification in the format 'port_name@mode_name'."""


PortMapMM: TypeAlias = dict[PortMode, int]
"""A mapping from multi-mode port-mode names to their matrix indices."""


PortCombinationMM: TypeAlias = tuple[PortMode, PortMode]
"""A pair of multi-mode port-mode names representing an S-parameter."""


SDictMM: TypeAlias = dict[PortCombinationMM, ComplexArray]
"""A sparse dictionary-based S-matrix representation.

A mapping from a port combination to an S-parameter or an array of S-parameters.

Examples:
    Creating an `SDictMM`:

    ```python
    sdict: sax.SDict = {
        ("in0", "out0"): 3.0,
    }
    ```

"""

SDenseMM: TypeAlias = tuple[ComplexArray, PortMapMM]
"""A dense S-matrix representation.

S-matrix (2D array) or multidimensional batched S-matrix (N+2)-D array with a port map.
If (N+2)-D array then the S-matrix dimensions are the last two.

Examples:
    Creating an `SDenseMM`:

    ```python
    Sd = jnp.arange(9, dtype=float).reshape(3, 3)
    port_map = {"in0": 0, "in1": 2, "out0": 1}
    sdense = Sd, port_map
    ```

"""

SCooMM: TypeAlias = tuple[IntArray1D, IntArray1D, ComplexArray, PortMapMM]
"""A sparse S-matrix in COO format (recommended for internal library use only).

An `SCoo` is a sparse matrix based representation of an S-matrix consisting of three
arrays and a port map. The three arrays represent the input port indices [`int`],
output port indices [`int`] and the S-matrix values [`ComplexFloat`] of the sparse
matrix. The port map maps a port name [`str`] to a port index [`int`].

Only these four arrays **together** and in this specific **order** are considered a
valid `SCoo` representation!

Examples:
    Creating an `SCooMM':

    ```python
    Si = jnp.arange(3, dtype=int)
    Sj = jnp.array([0, 1, 0], dtype=int)
    Sx = jnp.array([3.0, 4.0, 1.0])
    port_map = {"in0": 0, "in1": 2, "out0": 1}
    scoo: sax.SCoo = (Si, Sj, Sx, port_map)
    ```

Note:
    This representation is only recommended for internal library use. Please don't
    write user-facing code using this representation.

"""

STypeMM: TypeAlias = SDictMM | SCooMM | SDenseMM
"""Any S-Matrix type [SDict, SDense, SCOO]."""


def val_model(model: Any) -> ModelMM:
    """Validate a multi-mode SAX model function.

    Args:
        model: The model function to validate.

    Returns:
        The validated multi-mode model.

    Raises:
        TypeError: If validation fails.
    """
    return val_not_callable_annotated(val_sax_callable(model))


SDictModelMM: TypeAlias = Annotated[Callable[..., SDictMM], val(val_model)]
"""A keyword-only function that produces a multi-mode SDict S-matrix."""

SDenseModelMM: TypeAlias = Annotated[Callable[..., SDenseMM], val(val_model)]
"""A keyword-only function that produces a multi-mode SDense S-matrix."""


SCooModelMM: TypeAlias = Annotated[Callable[..., SCooMM], val(val_model)]
"""A keyword-only function that produces a multi-mode SCoo S-matrix."""


ModelMM: TypeAlias = Annotated[
    SDenseModelMM | SCooModelMM | SDictModelMM, val(val_model)  # type: ignore[reportInvalidTypeForm]
]
"""A keyword-only function that produces any multi-mode S-matrix type."""


def val_model_factory(model: Any) -> ModelFactoryMM:
    """Validate a multi-mode SAX model factory function.

    Args:
        model: The model factory function to validate.

    Returns:
        The validated multi-mode model factory.

    Raises:
        TypeError: If validation fails.
    """
    return val_callable_annotated(val_sax_callable(model))


SDictModelFactoryMM: TypeAlias = Annotated[
    Callable[..., SDictModelMM], val(val_model_factory)
]
"""A keyword-only function that produces a multi-mode SDict model."""


SDenseModelFactoryMM: TypeAlias = Annotated[
    Callable[..., SDenseModelMM], val(val_model_factory)
]
"""A keyword-only function that produces a multi-mode SDense model."""

SCooModelFactoryMM: TypeAlias = Annotated[
    Callable[..., SCooModelMM], val(val_model_factory)
]
"""A keyword-only function that produces a multi-mode SCoo model."""


ModelFactoryMM: TypeAlias = Annotated[
    SDictModelFactoryMM | SDenseModelFactoryMM | SCooModelFactoryMM,
    val(val_model_factory),
]
"""A keyword-only function that produces any multi-mode model."""

ModelsMM: TypeAlias = dict[Name, ModelMM]
"""A mapping from model names to multi-mode model functions."""



"""SAX Filipsson Gunnar Backend."""

from __future__ import annotations

from typing import Any

import jax
from jaxtyping import Array

import sax

__all__ = [
    "analyze_circuit_fg",
    "analyze_instances_fg",
    "evaluate_circuit_fg",
]


def analyze_instances_fg(
    instances: sax.Instances,
    models: sax.Models,
) -> dict[sax.InstanceName, sax.SDict]:
    """Analyze circuit instances for the Filipsson-Gunnar backend.

    Prepares instance S-matrices for the Filipsson-Gunnar backend by converting
    all component models to SDict format. This backend implements the classic
    Filipsson-Gunnar algorithm for S-matrix interconnection of multiports.

    Args:
        instances: Dictionary mapping instance names to instance definitions
            containing component names and settings.
        models: Dictionary mapping component names to their model functions.

    Returns:
        Dictionary mapping instance names to their S-matrices in SDict format.

    Note:
        The Filipsson-Gunnar algorithm is a systematic method for computing
        the overall S-matrix of interconnected multiport networks described in:
        Filipsson, Gunnar. "A new general computer algorithm for S-matrix
        calculation of interconnected multiports." 11th European Microwave
        Conference. IEEE, 1981.

    Example:
        ```python
        instances = {
            "wg1": {"component": "waveguide", "settings": {"length": 10.0}},
            "dc1": {"component": "coupler", "settings": {"coupling": 0.1}},
        }
        models = {"waveguide": waveguide_model, "coupler": coupler_model}
        analyzed = analyze_instances_fg(instances, models)
        ```
    """
    instances = sax.into[sax.Instances](instances)
    models = sax.into[sax.Models](models)
    model_names = set()
    for i in instances.values():
        model_names.add(i["component"])
    dummy_models = {k: sax.sdict(models[k]()) for k in model_names}
    dummy_instances = {}
    for k, i in instances.items():
        dummy_instances[k] = dummy_models[i["component"]]
    return dummy_instances


def analyze_circuit_fg(
    analyzed_instances: dict[str, sax.SDict],  # noqa: ARG001
    connections: sax.Connections,
    ports: sax.Ports,
) -> Any:  # noqa: ANN401
    """Analyze circuit topology for the Filipsson-Gunnar backend.

    Prepares the circuit connection information for the Filipsson-Gunnar backend
    evaluation. This implementation currently skips detailed analysis and passes
    the connection information directly to the evaluation phase.

    Args:
        analyzed_instances: Instance S-matrices from analyze_instances_fg.
            Not used in this analysis step but required for interface consistency.
        connections: Dictionary mapping instance ports to each other, defining
            internal circuit connections.
        ports: Dictionary mapping external port names to instance ports.

    Returns:
        Tuple containing connections and ports information for circuit evaluation.

    Example:
        ```python
        connections = {"wg1,out": "dc1,in1", "dc1,out1": "wg2,in"}
        ports = {"in": "wg1,in", "out": "wg2,out"}
        analyzed = analyze_circuit_fg(analyzed_instances, connections, ports)
        ```
    """
    return connections, ports  # skip analysis for now


def evaluate_circuit_fg(
    analyzed: Any,  # noqa: ANN401
    instances: dict[str, sax.SType],
) -> sax.SDict:
    """Evaluate circuit S-matrix using the Filipsson-Gunnar algorithm.

    Computes the overall circuit S-matrix by systematically interconnecting
    multiport networks using the Filipsson-Gunnar algorithm. This method
    iteratively applies equation 6 from the original paper to connect ports.

    The algorithm:
    1. Creates a block diagonal S-matrix from all component S-matrices
    2. Iteratively interconnects ports according to the connections
    3. Applies the Filipsson-Gunnar interconnection formula for each connection
    4. Extracts the final S-matrix for external ports

    Args:
        analyzed: Circuit analysis data from analyze_circuit_fg containing
            connections and ports information.
        instances: Dictionary mapping instance names to their evaluated S-matrices
            in any SAX format (will be converted to SDict).

    Returns:
        Circuit S-matrix in SDict format with external port combinations as keys.

    Note:
        The interconnection formula used is equation 6 from:
        Filipsson, Gunnar. "A new general computer algorithm for S-matrix
        calculation of interconnected multiports." 11th European Microwave
        Conference. IEEE, 1981.

    Example:
        ```python
        # Circuit analysis and instances
        analyzed = (connections, ports)
        instances = {
            "wg1": {("in", "out"): 0.95 * jnp.exp(1j * 0.1)},
            "dc1": {("in1", "out1"): 0.9, ("in1", "out2"): 0.1},
        }
        circuit_s = evaluate_circuit_fg(analyzed, instances)
        ```
    """
    connections, ports = analyzed

    # it's actually easier working w reverse:
    reversed_ports = {v: k for k, v in ports.items()}

    block_diag = {}
    for name, S in instances.items():
        block_diag.update(
            {
                (f"{name},{p1}", f"{name},{p2}"): v
                for (p1, p2), v in sax.sdict(S).items()
            }
        )

    sorted_connections = sorted(connections.items(), key=_connections_sort_key)
    all_connected_instances = {k: {k} for k in instances}

    for k, l in sorted_connections:
        name1, _ = k.split(",")
        name2, _ = l.split(",")

        connected_instances = (
            all_connected_instances[name1] | all_connected_instances[name2]
        )
        for name in connected_instances:
            all_connected_instances[name] = connected_instances

        current_ports = tuple(
            p
            for instance in connected_instances
            for p in set([p for p, _ in block_diag] + [p for _, p in block_diag])
            if p.startswith(f"{instance},")
        )

        block_diag.update(_interconnect_ports(block_diag, current_ports, k, l))

        for i, j in list(block_diag.keys()):
            is_connected = j in (k, l) or i in (k, l)
            is_in_output_ports = i in reversed_ports and j in reversed_ports
            if is_connected and not is_in_output_ports:
                del block_diag[
                    i, j
                ]  # we're no longer interested in these port combinations

    circuit_sdict: sax.SDict = {
        (reversed_ports[i], reversed_ports[j]): v
        for (i, j), v in block_diag.items()
        if i in reversed_ports and j in reversed_ports
    }
    return circuit_sdict


def _connections_sort_key(connection: tuple[str, str]) -> tuple[str, str]:
    """Sort key for sorting a connection dictionary."""
    part1, part2 = connection
    name1, _ = part1.split(",")
    name2, _ = part2.split(",")
    return (min(name1, name2), max(name1, name2))


def _interconnect_ports(
    block_diag: dict[tuple[str, str], Any],
    current_ports: tuple[str, ...],
    k: str,
    l: str,
) -> dict[tuple[str, str], Any]:
    """Interconnect two ports in a given model.

    > the interconnect algorithm is based on equation 6 of 'Filipsson, Gunnar.
    > "A new general computer algorithm for S-matrix calculation of interconnected
    > multiports." 11th European Microwave Conference. IEEE, 1981.'

    """
    current_block_diag = {}
    for i in current_ports:
        for j in current_ports:
            vij = _calculate_interconnected_value(
                vij=block_diag.get((i, j), 0.0),
                vik=block_diag.get((i, k), 0.0),
                vil=block_diag.get((i, l), 0.0),
                vkj=block_diag.get((k, j), 0.0),
                vkk=block_diag.get((k, k), 0.0),
                vkl=block_diag.get((k, l), 0.0),
                vlj=block_diag.get((l, j), 0.0),
                vlk=block_diag.get((l, k), 0.0),
                vll=block_diag.get((l, l), 0.0),
            )
            current_block_diag[i, j] = vij
    return current_block_diag


@jax.jit
def _calculate_interconnected_value(
    vij: Array,
    vik: Array,
    vil: Array,
    vkj: Array,
    vkk: Array,
    vkl: Array,
    vlj: Array,
    vlk: Array,
    vll: Array,
) -> Array:
    """Calculate an interconnected S-parameter value.

    > The interconnect algorithm is based on equation 6 in the paper below
    > Filipsson, Gunnar. "A new general computer algorithm for S-matrix calculation
    > of interconnected multiports." 11th European Microwave Conference. IEEE, 1981.

    """
    result = vij + (
        vkj * vil * (1 - vlk)
        + vlj * vik * (1 - vkl)
        + vkj * vll * vik
        + vlj * vkk * vil
    ) / ((1 - vkl) * (1 - vlk) - vkk * vll)
    return result



"""SAX Probe Models."""

from __future__ import annotations

import jax.numpy as jnp
from pydantic import validate_call

import sax

__all__ = ["ideal_probe"]


@validate_call
def ideal_probe(wl: sax.FloatArrayLike = sax.WL_C) -> sax.SDict:
    """Ideal 4-port measurement probe with 100% transmission and 100% tap coupling.

    This is an unphysical component designed for debugging and measurement purposes.
    It intercepts a connection and provides access to both forward and backward
    traveling waves without affecting the signal propagation.

    ```
              in ─────────────────── out
               │                     │
               │     (ideal tap)     │
               │                     │
             tap_bwd               tap_fwd
    ```

    The probe has the following behavior:
    - Full transmission from `in` to `out` (and vice versa)
    - Forward tap (`tap_fwd`) copies the signal traveling from `in` toward `out`
    - Backward tap (`tap_bwd`) copies the signal traveling from `out` toward `in`
    - No reflections at any port
    - No cross-coupling between tap ports

    Note:
        This S-matrix is NOT unitary (violates energy conservation). This is
        intentional—it's a measurement tool, not a physical device.

    Args:
        wl: Wavelength in micrometers. Defaults to 1.55 μm.

    Returns:
        S-parameter dictionary for the ideal probe.

    Example:
        Probes are typically not used directly, but via the `probes` argument
        to `sax.circuit()`:

        ```python
        circuit_fn, info = sax.circuit(
            netlist,
            models=models,
            probes={"mid": "wg1,out"},
        )
        # Circuit now has additional ports: mid_fwd, mid_bwd
        ```
    """
    one = jnp.ones_like(jnp.asarray(wl))
    zero = jnp.zeros_like(jnp.asarray(wl))
    return {
        # Through path: full transmission
        ("in", "out"): one,
        ("out", "in"): one,
        # Forward tap: copies signal from in→out direction
        ("in", "tap_fwd"): one,
        ("tap_fwd", "in"): one,
        # Backward tap: copies signal from out→in direction
        ("out", "tap_bwd"): one,
        ("tap_bwd", "out"): one,
        # No cross-coupling between taps
        ("tap_fwd", "tap_bwd"): zero,
        ("tap_bwd", "tap_fwd"): zero,
        # No coupling from taps to opposite main port
        ("tap_fwd", "out"): zero,
        ("tap_bwd", "in"): zero,
        # No reflections
        ("in", "in"): zero,
        ("out", "out"): zero,
        ("tap_fwd", "tap_fwd"): zero,
        ("tap_bwd", "tap_bwd"): zero,
    }



"""Tests for the probes feature."""

import pytest

import sax


def test_ideal_probe_model() -> None:
    """Test that the ideal_probe model returns the expected S-matrix."""
    s = sax.models.ideal_probe(wl=1.55)

    # Check all expected ports exist
    expected_ports = {"in", "out", "tap_fwd", "tap_bwd"}
    ports = sax.get_ports(s)
    assert set(ports) == expected_ports

    # Check through path has full transmission
    assert float(s[("in", "out")]) == 1.0
    assert float(s[("out", "in")]) == 1.0

    # Check tap coupling
    assert float(s[("in", "tap_fwd")]) == 1.0
    assert float(s[("tap_fwd", "in")]) == 1.0
    assert float(s[("out", "tap_bwd")]) == 1.0
    assert float(s[("tap_bwd", "out")]) == 1.0

    # Check no cross-coupling between taps
    assert float(s[("tap_fwd", "tap_bwd")]) == 0.0
    assert float(s[("tap_bwd", "tap_fwd")]) == 0.0

    # Check no reflections
    assert float(s[("in", "in")]) == 0.0
    assert float(s[("out", "out")]) == 0.0
    assert float(s[("tap_fwd", "tap_fwd")]) == 0.0
    assert float(s[("tap_bwd", "tap_bwd")]) == 0.0


def test_basic_probe_insertion() -> None:
    """Test basic probe insertion on a simple 2-waveguide circuit."""
    netlist = {
        "instances": {
            "wg1": "waveguide",
            "wg2": "waveguide",
        },
        "connections": {
            "wg1,out0": "wg2,in0",
        },
        "ports": {
            "in": "wg1,in0",
            "out": "wg2,out0",
        },
    }

    models = {
        "waveguide": sax.models.straight,
    }

    # Without probes
    circuit_no_probe, _ = sax.circuit(netlist, models)
    result_no_probe = circuit_no_probe()
    ports_no_probe = sax.get_ports(result_no_probe)
    assert set(ports_no_probe) == {"in", "out"}

    # With probe
    circuit_with_probe, _ = sax.circuit(
        netlist,
        models,
        probes={"mid": "wg1,out0"},
    )
    result_with_probe = circuit_with_probe()
    ports_with_probe = sax.get_ports(result_with_probe)
    assert set(ports_with_probe) == {"in", "out", "mid_fwd", "mid_bwd"}


def test_probe_on_right_side_of_connection() -> None:
    """Test that probes work when specified on the right side of a connection."""
    netlist = {
        "instances": {
            "wg1": "waveguide",
            "wg2": "waveguide",
        },
        "connections": {
            "wg1,out0": "wg2,in0",
        },
        "ports": {
            "in": "wg1,in0",
            "out": "wg2,out0",
        },
    }

    models = {
        "waveguide": sax.models.straight,
    }

    # Probe on right side of connection (wg2,in0 instead of wg1,out0)
    circuit_fn, _ = sax.circuit(
        netlist,
        models,
        probes={"mid": "wg2,in0"},
    )
    result = circuit_fn()
    ports = sax.get_ports(result)
    assert set(ports) == {"in", "out", "mid_fwd", "mid_bwd"}


def test_probe_fwd_direction_left_side() -> None:
    """Test _fwd captures signal flowing INTO user-specified port (left side)."""
    netlist = {
        "instances": {
            "wg1": "waveguide",
            "wg2": "waveguide",
        },
        "connections": {
            "wg1,out0": "wg2,in0",
        },
        "ports": {
            "in": "wg1,in0",
            "out": "wg2,out0",
        },
    }

    models = {
        "waveguide": sax.models.straight,
    }

    # Probe at wg1,out0 (left side of connection)
    # _fwd should capture signal flowing INTO wg1,out0 (i.e., from wg2 back to wg1)
    # _bwd should capture signal flowing OUT of wg1,out0 (i.e., from wg1 to wg2)
    circuit_fn, _ = sax.circuit(
        netlist,
        models,
        probes={"mid": "wg1,out0"},
    )
    result = circuit_fn()

    # Signal from "in" goes through wg1 and exits at wg1,out0
    # This should appear at mid_bwd (signal flowing OUT of the probed port)
    assert ("in", "mid_bwd") in result

    # Signal from "out" goes back through wg2 and into wg1,out0
    # This should appear at mid_fwd (signal flowing INTO the probed port)
    assert ("out", "mid_fwd") in result


def test_probe_fwd_direction_right_side() -> None:
    """Test _fwd captures signal flowing INTO user-specified port (right side)."""
    netlist = {
        "instances": {
            "wg1": "waveguide",
            "wg2": "waveguide",
        },
        "connections": {
            "wg1,out0": "wg2,in0",
        },
        "ports": {
            "in": "wg1,in0",
            "out": "wg2,out0",
        },
    }

    models = {
        "waveguide": sax.models.straight,
    }

    # Probe at wg2,in0 (right side of connection)
    # _fwd should capture signal flowing INTO wg2,in0 (i.e., from wg1 to wg2)
    # _bwd should capture signal flowing OUT of wg2,in0 (i.e., from wg2 back to wg1)
    circuit_fn, _ = sax.circuit(
        netlist,
        models,
        probes={"mid": "wg2,in0"},
    )
    result = circuit_fn()

    # Signal from "in" goes through wg1 and into wg2,in0
    # This should appear at mid_fwd (signal flowing INTO the probed port)
    assert ("in", "mid_fwd") in result

    # Signal from "out" goes back through wg2 and exits at wg2,in0
    # This should appear at mid_bwd (signal flowing OUT of the probed port)
    assert ("out", "mid_bwd") in result


def test_multiple_probes() -> None:
    """Test multiple probes on different connections."""
    netlist = {
        "instances": {
            "wg1": "waveguide",
            "wg2": "waveguide",
            "wg3": "waveguide",
        },
        "connections": {
            "wg1,out0": "wg2,in0",
            "wg2,out0": "wg3,in0",
        },
        "ports": {
            "in": "wg1,in0",
            "out": "wg3,out0",
        },
    }

    models = {
        "waveguide": sax.models.straight,
    }

    circuit_fn, _ = sax.circuit(
        netlist,
        models,
        probes={
            "probe1": "wg1,out0",
            "probe2": "wg2,out0",
        },
    )
    result = circuit_fn()
    ports = sax.get_ports(result)
    expected_ports = {
        "in",
        "out",
        "probe1_fwd",
        "probe1_bwd",
        "probe2_fwd",
        "probe2_bwd",
    }
    assert set(ports) == expected_ports


def test_probe_in_mzi_circuit() -> None:
    """Test probe in a more complex MZI circuit."""
    netlist = {
        "instances": {
            "lft": "coupler",
            "top": "waveguide",
            "btm": "waveguide",
            "rgt": "coupler",
        },
        "connections": {
            "lft,out0": "btm,in0",
            "btm,out0": "rgt,in0",
            "lft,out1": "top,in0",
            "top,out0": "rgt,in1",
        },
        "ports": {
            "in0": "lft,in0",
            "in1": "lft,in1",
            "out0": "rgt,out0",
            "out1": "rgt,out1",
        },
    }

    models = {
        "coupler": sax.models.coupler_ideal,
        "waveguide": sax.models.straight,
    }

    # Probe on top arm
    circuit_fn, _ = sax.circuit(
        netlist,
        models,
        probes={"top_arm": "top,in0"},
    )
    result = circuit_fn()
    ports = sax.get_ports(result)
    expected_ports = {"in0", "in1", "out0", "out1", "top_arm_fwd", "top_arm_bwd"}
    assert set(ports) == expected_ports


def test_probe_values_match_transmission() -> None:
    """Test that probe values match expected transmission through the circuit."""
    netlist = {
        "instances": {
            "wg1": "waveguide",
            "wg2": "waveguide",
        },
        "connections": {
            "wg1,out0": "wg2,in0",
        },
        "ports": {
            "in": "wg1,in0",
            "out": "wg2,out0",
        },
    }

    models = {
        "waveguide": sax.models.straight,
    }

    circuit_fn, _ = sax.circuit(
        netlist,
        models,
        probes={"mid": "wg1,out0"},
    )

    # Evaluate at a specific wavelength
    result = circuit_fn(wl=1.55)

    # The forward probe should capture signal going from wg1 to wg2
    # The transmission in->out should still work
    assert ("in", "out") in result
    assert ("in", "mid_fwd") in result
    assert ("out", "mid_bwd") in result


def test_probe_on_unconnected_port() -> None:
    """Test that probes on unconnected ports create a fwd-only port alias."""
    netlist = {
        "instances": {
            "wg1": "waveguide",
            "wg2": "waveguide",
            "wg3": "waveguide",
        },
        "connections": {
            "wg1,out0": "wg2,in0",
            "wg2,out0": "wg3,in0",
        },
        "ports": {
            "in": "wg1,in0",
            "out": "wg3,out0",
        },
    }

    models = {
        "waveguide": sax.models.straight,
    }

    # wg1,in0 is not part of any connection (it's a top-level port).
    # Probing it should create only tap_fwd as an alias.
    circuit_fn, _ = sax.circuit(
        netlist,
        models,
        probes={"tap": "wg1,in0"},
    )
    result = circuit_fn()
    ports = sax.get_ports(result)
    # Only tap_fwd should be created (no tap_bwd for unconnected port)
    assert "tap_fwd" in ports
    assert "tap_bwd" not in ports
    assert set(ports) == {"in", "out", "tap_fwd"}


def test_probe_error_on_port_conflict() -> None:
    """Test that probes raise error when port names would conflict."""
    netlist = {
        "instances": {
            "wg1": "waveguide",
            "wg2": "waveguide",
        },
        "connections": {
            "wg1,out0": "wg2,in0",
        },
        "ports": {
            "in": "wg1,in0",
            "out": "wg2,out0",
            "mid_fwd": "wg2,out0",  # Conflict with probe port
        },
    }

    models = {
        "waveguide": sax.models.straight,
    }

    with pytest.raises(ValueError, match="conflict with existing ports"):
        sax.circuit(
            netlist,
            models,
            probes={"mid": "wg1,out0"},
        )


def test_probe_error_on_instance_conflict() -> None:
    """Test that probes raise error when instance name would conflict."""
    netlist = {
        "instances": {
            "wg1": "waveguide",
            "wg2": "waveguide",
            "_probe_mid": "waveguide",  # Conflict with probe instance name
        },
        "connections": {
            "wg1,out0": "wg2,in0",
        },
        "ports": {
            "in": "wg1,in0",
            "out": "wg2,out0",
        },
    }

    models = {
        "waveguide": sax.models.straight,
    }

    with pytest.raises(ValueError, match="conflicts with an existing instance"):
        sax.circuit(
            netlist,
            models,
            probes={"mid": "wg1,out0"},
        )


def test_empty_probes_dict() -> None:
    """Test that empty probes dict is a no-op."""
    netlist = {
        "instances": {
            "wg1": "waveguide",
            "wg2": "waveguide",
        },
        "connections": {
            "wg1,out0": "wg2,in0",
        },
        "ports": {
            "in": "wg1,in0",
            "out": "wg2,out0",
        },
    }

    models = {
        "waveguide": sax.models.straight,
    }

    circuit_fn, _ = sax.circuit(netlist, models, probes={})
    result = circuit_fn()
    ports = sax.get_ports(result)
    assert set(ports) == {"in", "out"}


def test_probe_with_none() -> None:
    """Test that probes=None is a no-op."""
    netlist = {
        "instances": {
            "wg1": "waveguide",
            "wg2": "waveguide",
        },
        "connections": {
            "wg1,out0": "wg2,in0",
        },
        "ports": {
            "in": "wg1,in0",
            "out": "wg2,out0",
        },
    }

    models = {
        "waveguide": sax.models.straight,
    }

    circuit_fn, _ = sax.circuit(netlist, models, probes=None)
    result = circuit_fn()
    ports = sax.get_ports(result)
    assert set(ports) == {"in", "out"}


if __name__ == "__main__":
    test_ideal_probe_model()
    test_basic_probe_insertion()
    test_probe_on_right_side_of_connection()
    test_multiple_probes()
    test_probe_in_mzi_circuit()
    test_probe_values_match_transmission()
    test_probe_on_unconnected_port()
    test_probe_error_on_port_conflict()
    test_probe_error_on_instance_conflict()
    test_empty_probes_dict()
    test_probe_with_none()
    print("All tests passed!")



from functools import partial

import sax

SAMPLE_NETLIST = {
    "instances": {
        "lft": "coupler",
        "top": "waveguide",
        "btm": "waveguide",
        "rgt": "coupler",
    },
    "connections": {
        "lft,out0": "btm,in0",
        "btm,out0": "rgt,in0",
        "lft,out1": "top,in0",
        "top,out0": "rgt,in1",
    },
    "ports": {
        "in0": "lft,in0",
        "in1": "lft,in1",
        "out0": "rgt,out0",
        "out1": "rgt,out1",
    },
}

SAMPLE_NETLIST2 = {
    "instances": {
        "lft": sax.models.coupler_ideal,
        "top": partial(sax.models.straight, length=10.0),
        "btm": partial(sax.models.straight, length=30.0),
        "rgt": sax.models.coupler_ideal,
    },
    "connections": {
        "lft,out0": "btm,in0",
        "btm,out0": "rgt,in0",
        "lft,out1": "top,in0",
        "top,out0": "rgt,in1",
    },
    "ports": {
        "in0": "lft,in0",
        "in1": "lft,in1",
        "out0": "rgt,out0",
        "out1": "rgt,out1",
    },
}


def test_netlist_function_from_dict() -> None:
    net = sax.netlist(SAMPLE_NETLIST)
    assert "top_level" in net


def test_netlist_using_instance_functions() -> None:
    net = sax.netlist(SAMPLE_NETLIST2)
    assert "top_level" in net



"""Parser for Lumerical S-parameter files.

Based on the simphony lumerical parser.
"""

from __future__ import annotations

import re
import warnings
from hashlib import md5
from pathlib import Path
from tempfile import gettempdir
from textwrap import dedent
from typing import cast, overload

import numpy as np
import pandas as pd
from lark import Lark, Token, Transformer, Tree, v_args
from numpy.typing import NDArray
from typing_extensions import TypedDict

import sax

from .touchstone import _validate_columns


def parse_lumerical_dat(
    content_or_filename: str | Path | sax.IOLike,
    *,
    convert_f_to_wl: bool = False,
) -> pd.DataFrame:
    """Load S-parameters from a Lumerical .dat or .sparam file.

    Args:
        content_or_filename: Content as string (if contains newlines), file path,
            or file-like object with read() method.
        convert_f_to_wl: If True, convert frequency to wavelength.

    Returns:
        a pandas DataFrame with the S-parameters.

    Note:
        This function was adapted from the `simphony` load_sparams function.
    """
    msg = (
        "The `parse_lumerical_dat` function is experimental. "
        "If you encounter any issues, "
        "Please file a bug report here: https://github.com/flaport/sax/issues ."
    )
    warnings.warn(msg, stacklevel=2, category=sax.ExperimentalWarning)
    content = sax.read(content_or_filename)
    _tree, df = cast(tuple[Tree, pd.DataFrame], _parser.parse(content))
    df = df.rename(columns={"phase": "phi"})
    x = "f"
    if convert_f_to_wl:
        x = "wl"
        df["wl"] = sax.C_UM_S / df.pop("f")
    df["amp"] = df.pop("mag")
    return cast(
        pd.DataFrame,
        df[[x, "port_in", "port_out", "mode_in", "mode_out", "amp", "phi"]],
    )


@overload
def write_lumerical_dat(df: pd.DataFrame, path: None) -> str: ...


@overload
def write_lumerical_dat(df: pd.DataFrame, path: str | Path) -> Path: ...


def write_lumerical_dat(
    df: pd.DataFrame,
    path: str | Path | None = None,
) -> Path | str:
    """Write S-parameters to a Lumerical .dat or .sparam file.

    Args:
        df: DataFrame in tidy format containing S-parameters.
        path: Path to the output file. If None, the string content is returned

    Returns:
        Path to the written file or string content if path is None.
    """
    temppath = None
    if path is None:
        temppath = path = (
            Path(gettempdir()).resolve()
            / "sax"
            / f"write_lumerical_{md5(df.to_numpy().tobytes()).hexdigest()}.dat"
        )
    in_amp_phi_format, in_wl_format = _validate_columns(df)
    modes = {*df["mode_in"], *df["mode_out"]}
    if in_amp_phi_format:
        df["s"] = df["amp"] * np.exp(1j * df["phi"])
        df = df.drop(columns=["amp", "phi"])
    else:
        df["s"] = df["re"] + 1j * df["im"]
        df = df.drop(columns=["re", "im"])
    if len(modes) > 1:
        df["port_in"] = [
            f"{p}@{m}" for p, m in zip(df["port_in"], df["mode_in"], strict=True)
        ]
        df["port_out"] = [
            f"{p}@{m}" for p, m in zip(df["port_out"], df["mode_out"], strict=True)
        ]
    df = df.drop(columns=["mode_in", "mode_out"])
    if in_wl_format:
        df["f"] = sax.C_UM_S / df.pop("wl")
    df = cast(pd.DataFrame, df[["f", "port_in", "port_out", "s"]])
    xarr = sax.to_xarray(df, target_names=["s"])
    freq = xarr.coords["f"].to_numpy()
    sparams = xarr.to_numpy()[:, :, :, 0]
    buf = Path(path).open("a")  # noqa: SIM115
    try:
        d = sparams.shape[1]
        for in_ in range(d):
            for out in range(d):
                sp = sparams[:, in_, out]
                temp = np.vstack((freq, np.abs(sp), np.unwrap(np.angle(sp)))).T
                header = f'("port {out + 1}", "TE", 1, "port {in_ + 1}", 1, "transmission")\n'
                header += f"{temp.shape}"
                np.savetxt(buf, temp, header=header, comments="")
    finally:
        buf.close()

    if temppath:
        return temppath.read_text()
    return Path(path).resolve()


class _SparamsTransformer(Transformer):
    @v_args(inline=True)
    def start(self, header: str, *datablocks: pd.DataFrame) -> tuple[str, pd.DataFrame]:
        data = pd.concat(datablocks, ignore_index=True)
        return header, data

    @v_args(inline=True)
    def datablock(
        self, ports: _PortsDict, shape: tuple[int, int], values: NDArray[np.float64]
    ) -> pd.DataFrame:
        sweepparams = ports["sweepparams"] or []
        columns = np.array([*sweepparams, "f", "mag", "phase"])
        df = pd.DataFrame(values, columns=columns)
        rows, _ = shape
        if ports["groupdelay"] is not None:
            groupshift = (
                2
                * np.pi
                * float(ports["groupdelay"])
                * (df.loc[:, "f"] - df.loc[int(rows / 2), "f"])
            )
            df.loc[:, "phase"] += groupshift
        df.loc[:, "port_out"] = ports["port_out"]
        df.loc[:, "port_in"] = ports["port_in"]
        df.loc[:, "mode_out"] = ports["mode_out"]
        df.loc[:, "mode_in"] = ports["mode_in"]
        return df

    @v_args(inline=True)
    def ports(
        self,
        port_out: str,
        mode_type: str,
        mode_out: int,
        port_in: str,
        mode_in: int,
        valuetype: str,
        groupdelay: str | None = None,
        sweepparams: str | None = None,
    ) -> _PortsDict:
        return cast(
            _PortsDict,
            {
                "port_out": port_out,
                "mode_type": mode_type,
                "mode_out": mode_out,
                "port_in": port_in,
                "mode_in": mode_in,
                "valuetype": valuetype,
                "groupdelay": groupdelay,
                "sweepparams": sweepparams,
            },
        )

    def shape(self, args: list[Token]) -> tuple[int, int]:
        int_args = [int(arg) for arg in args]
        if len(int_args) != 2:
            msg = f"Expected 2 arguments for shape, got {len(int_args)}"
            raise ValueError(msg)
        return (int_args[0], int_args[1])

    def values(self, args: list[NDArray[np.float64]]) -> np.ndarray:
        return np.array(args)

    def row(self, args: list[Token]) -> np.ndarray:
        return np.array([float(arg) for arg in args])

    @v_args(inline=True)
    def port(self, port: Token) -> str:
        return _clean_string(_destring(port))

    @v_args(inline=True)
    def modeid(self, mid: Token) -> int:
        if isinstance(mid, int):
            return mid
        msg = f"Mode ID '{mid}' is not supported, contact the developers."
        raise ValueError(msg)

    @v_args(inline=True)
    def groupdelay(self, value: Token) -> float:
        return float(value)

    @v_args(inline=True)
    def sweepparams(self, params: Token) -> list[str]:
        return _destring(str(params)).split(";")

    def MODE(self, args: Token) -> str:
        return _destring(str(args))

    def VALUETYPE(self, args: Token) -> str:
        return _destring(str(args))

    def INT(self, args: Token) -> int:
        return int(args)

    def STRING(self, args: Token) -> str:
        return str(args)


def _clean_string(s: str, dot: str = "p", minus: str = "m", other: str = "_") -> str:
    s = s.strip()
    s = s.replace(".", dot)  # dot
    s = s.replace("-", minus)  # minus
    s = re.sub("[^0-9a-zA-Z_]", other, s)
    if s[0] in "0123456789":
        s = "_" + s
    if not s.isidentifier():
        msg = f"failed to clean string to a valid python identifier: {s}"
        raise ValueError(msg)
    return s


def _destring(string: str) -> str:
    return string.replace("'", "").replace('"', "")


_sparams_grammar = dedent(r"""
    ?start: _EOL* [header] datablock+ _EOL*

    header: option1 | option2
    option1: "[" INT "," INT "]" _EOL
    option2: ("[" port "," position "]" _EOL)+

    datablock: ports shape values

    ports: "(" port "," MODE "," modeid "," port "," modeid "," VALUETYPE ["," groupdelay] ["," sweepparams] ")" _EOL
    shape: "(" INT "," INT ")" _EOL
    values: row+

    row: (SIGNED_NUMBER)+ _EOL
    position: ("'" [SIDE] "'") | ("\"" [SIDE] "\"")
    port: STRING
    modeid: INT
    VALUETYPE: STRING
    groupdelay: SIGNED_NUMBER
    sweepparams: STRING

    SIDE: "TOP" | "BOTTOM" | "LEFT" | "RIGHT"
    MODE: ("'" [POL] "'") | ("\"" [POL] "\"") | STRING
    POL: "TE" | "TM"
    _EOL: NEWLINE
    STRING: ("'" /[^']*?/ _STRING_ESC_INNER "'") | ("\"" /[^\"]*?/ _STRING_ESC_INNER "\"")

    %import common._STRING_ESC_INNER
    %import common.SIGNED_NUMBER
    %import common.NUMBER
    %import common.INT
    %import common.WS_INLINE
    %import common.WS
    %import common.NEWLINE
    %ignore WS_INLINE
""")

_parser = Lark(
    _sparams_grammar, start="start", parser="lalr", transformer=_SparamsTransformer()
)


class _PortsDict(TypedDict):
    port_out: str
    mode_type: str
    mode_out: int
    port_in: str
    mode_in: int
    valuetype: str
    groupdelay: float | None
    sweepparams: str | None



"""SAX Default Models."""

from __future__ import annotations

import jax
import jax.numpy as jnp
from pydantic import validate_call

import sax


@jax.jit
@validate_call
def splitter_ideal(
    *,
    wl: sax.FloatArrayLike = sax.WL_C,
    coupling: sax.FloatArrayLike = 0.5,
) -> sax.SDict:
    r"""Ideal 1x2 power splitter model.

    ```{svgbob}
                    .--------- out0
                   /           o2
                  /   .-------
    in0 ---------'   /
     o1             (
        ---------.   \
                  \   '------- out1
                   \           o3
                    '---------
    ```

    Args:
        wl: Wavelength in micrometers.
        coupling: Power coupling ratio between 0 and 1.

    Returns:
        S-matrix dictionary containing the complex-valued cross/thru coefficients.

    Examples:
        Ideal 1x2 splitter:

        ```python
        # mkdocs: render
        import matplotlib.pyplot as plt
        import numpy as np
        import sax

        sax.set_port_naming_strategy("optical")

        wl = sax.wl_c()
        s = sax.models.splitter_ideal(wl=wl, coupling=0.3)
        thru = np.abs(s[("o1", "o3")]) ** 2
        cross = np.abs(s[("o1", "o2")]) ** 2
        plt.figure()
        plt.plot(wl, thru, label="thru")
        plt.plot(wl, cross, label="cross")
        plt.xlabel("Wavelength [μm]")
        plt.ylabel("Power")
        plt.legend()
        ```
    """
    one = jnp.ones_like(jnp.asarray(wl))
    kappa = jnp.asarray(coupling**0.5) * one
    tau = jnp.asarray((1 - coupling) ** 0.5) * one

    p = sax.PortNamer(num_inputs=1, num_outputs=2)
    return sax.reciprocal(
        {
            (p.in0, p.out0): tau,
            (p.in0, p.out1): kappa,
        },
    )



"""Common data file parsers."""

from .lumerical import parse_lumerical_dat, write_lumerical_dat
from .touchstone import parse_touchstone, write_touchstone

__all__ = [
    "parse_lumerical_dat",
    "parse_touchstone",
    "write_lumerical_dat",
    "write_touchstone",
]



"""Netlist utilities."""

from __future__ import annotations

import warnings
from copy import deepcopy
from typing import cast, overload

import networkx as nx
from natsort import natsorted

import sax

__all__ = [  # noqa: RUF022
    "netlist",
    "flatten_netlist",
    "expand_probes",
]


def netlist(
    netlist: sax.AnyNetlist,
    *,
    top_level_name: str = "top_level",
) -> sax.RecursiveNetlist:
    """Convert a netlist to recursive netlist format.

    Ensures that the input netlist is in recursive netlist format, which is a
    dictionary mapping component names to their netlists. If a top-level circuit
    with the specified name exists, it will be promoted to the top level.

    Args:
        netlist: Input netlist in any format (flat netlist or recursive netlist).
        top_level_name: Name to use for the top-level circuit when converting
            from flat netlist format. Defaults to "top_level".

    Returns:
        Recursive netlist dictionary with component name keys.

    Raises:
        ValueError: If the input netlist format is invalid.

    Example:
        ```python
        # Convert flat netlist to recursive format
        flat_net = {
            "instances": {"wg1": {"component": "waveguide"}},
            "ports": {"in": "wg1,in", "out": "wg1,out"},
        }
        rec_net = netlist(flat_net, top_level_name="my_circuit")
        # Result: {"my_circuit": flat_net}
        ```
    """
    if not isinstance(netlist, dict):
        msg = (
            "Invalid argument for `netlist`. "
            "Expected type: dict | Netlist | RecursiveNetlist. "
            f"Got: {type(netlist)}."
        )
        raise TypeError(msg)

    if "instances" in netlist:
        return {top_level_name: cast(sax.Netlist, netlist)}

    recnet = cast(sax.RecursiveNetlist, netlist)
    top_level = recnet.get(top_level_name, None)
    if top_level is None:
        return recnet
    return {top_level_name: top_level, **recnet}


def flatten_netlist(recnet: sax.RecursiveNetlist, sep: str = "~") -> sax.Netlist:
    """Flatten a recursive netlist into a single flat netlist.

    Converts a hierarchical (recursive) netlist into a single flat netlist by
    inlining all sub-circuits. Instance names are prefixed to avoid conflicts.

    Args:
        recnet: Recursive netlist to flatten.
        sep: Separator used for hierarchical instance naming. Defaults to "~".

    Returns:
        Single flat netlist with all hierarchies inlined.

    Example:
        ```python
        # Flatten a hierarchical netlist
        recnet = {
            "top": {
                "instances": {"sub1": {"component": "subcircuit"}},
                "ports": {"in": "sub1,in"},
            },
            "subcircuit": {
                "instances": {"wg1": {"component": "waveguide"}},
                "ports": {"in": "wg1,in"},
            },
        }
        flat = flatten_netlist(recnet)
        # Result has instances like "sub1~wg1" for the flattened hierarchy
        ```
    """
    first_name = next(iter(recnet.keys()))
    net = deepcopy(recnet[first_name])
    _flatten_netlist_into(recnet, net, sep)
    return net


@overload
def remove_unused_instances(netlist: sax.Netlist) -> sax.Netlist: ...


@overload
def remove_unused_instances(netlist: sax.RecursiveNetlist) -> sax.RecursiveNetlist: ...


def remove_unused_instances(netlist: sax.AnyNetlist) -> sax.AnyNetlist:
    if "instances" in netlist:
        net = cast(sax.Netlist, deepcopy(netlist))
        names = _get_nodes_to_remove(_get_connectivity_graph(net), net)
        _remove_instances(net, names)
        _remove_connections(net, names)
        _remove_ports(net, names)
        return net

    recnet = cast(sax.RecursiveNetlist, {**netlist})
    for name, flat_netlist in recnet.items():
        recnet[name] = remove_unused_instances(flat_netlist)

    return recnet


@overload
def rename_instances(
    netlist: sax.Netlist,
    mapping: dict[sax.InstanceName, sax.InstanceName],
) -> sax.Netlist: ...


@overload
def rename_instances(
    netlist: sax.RecursiveNetlist,
    mapping: dict[sax.InstanceName, sax.InstanceName],
) -> sax.RecursiveNetlist: ...


def rename_instances(
    netlist: sax.AnyNetlist,
    mapping: dict[sax.InstanceName, sax.InstanceName],
) -> sax.AnyNetlist:
    if (recnet := sax.try_into[sax.RecursiveNetlist](netlist)) is not None:
        return {k: rename_instances(v, mapping) for k, v in recnet.items()}

    # it's a sax.Netlist now:
    net: sax.Netlist = sax.into[sax.Netlist](netlist)
    new: sax.Netlist = {"instances": {}, "ports": {}}
    inverse_mapping = {v: k for k, v in mapping.items()}
    if len(inverse_mapping) != len(mapping):
        msg = "Duplicate names to map onto found."
        raise ValueError(msg)
    new["instances"] = {mapping.get(k, k): v for k, v in net["instances"].items()}
    new["connections"] = {}
    for ip1, ip2 in net.get("connections", {}).items():
        i1, p1 = ip1.split(",")
        i2, p2 = ip2.split(",")
        i1 = mapping.get(i1, i1)
        i2 = mapping.get(i2, i2)
        new["connections"][f"{i1},{p1}"] = f"{i2},{p2}"
    new["ports"] = {}
    for q, ip in net.get("ports", {}).items():
        i, p = ip.split(",")
        i = mapping.get(i, i)
        new["ports"][q] = f"{i},{p}"

    new["placements"] = {
        mapping.get(k, k): v for k, v in net.get("placements", {}).items()
    }
    return {**net, **new}


@overload
def rename_models(
    netlist: sax.Netlist,
    mapping: dict[sax.Name, sax.Name],
) -> sax.Netlist: ...


@overload
def rename_models(
    netlist: sax.RecursiveNetlist,
    mapping: dict[sax.Name, sax.Name],
) -> sax.RecursiveNetlist: ...


def rename_models(
    netlist: sax.AnyNetlist,
    mapping: dict[sax.Name, sax.Name],
) -> sax.AnyNetlist:
    if (recnet := sax.try_into[sax.RecursiveNetlist](netlist)) is not None:
        return {k: rename_models(v, mapping) for k, v in recnet.items()}

    # it's a sax.Netlist now:
    net: sax.Netlist = deepcopy(sax.into[sax.Netlist](netlist))
    inverse_mapping = {v: k for k, v in mapping.items()}
    if len(inverse_mapping) != len(mapping):
        msg = "Duplicate names to map onto found."
        raise ValueError(msg)

    for inst in net["instances"].values():
        inst["component"] = mapping.get(inst["component"], inst["component"])
    return net


def _remove_ports(net: sax.Netlist, names: set[sax.InstanceName]) -> None:
    for pname, conn in list(net.get("ports", {}).items()):
        name, _ = conn.split(",")
        if name in names and pname in net.get("ports", {}):
            del net.get("ports", {})[pname]


def _remove_connections(net: sax.Netlist, names: set[sax.InstanceName]) -> None:
    net["connections"] = {**net.get("connections", {})}
    for conn1, conn2 in list(net["connections"].items()):
        for conn in [conn1, conn2]:
            name, _ = conn.split(",")
            if name in names and conn1 in net["connections"]:
                del net["connections"][conn1]
    if not net["connections"]:
        del net["connections"]


def _remove_instances(net: sax.Netlist, names: set[sax.InstanceName]) -> None:
    for name in names:
        if name in net["instances"]:
            del net["instances"][name]


def _get_nodes_to_remove(graph: nx.Graph, netlist: sax.Netlist) -> set[str]:
    nodes = set()
    for port in netlist.get("ports", {}):
        nodes |= nx.descendants(graph, port)
    return set(graph.nodes) - nodes


def _get_connectivity_netlist(netlist: sax.Netlist) -> dict:
    return {
        "instances": natsorted(netlist["instances"]),
        "connections": [
            (c1.split(",")[0], c2.split(",")[0])
            for c1, c2 in netlist.get("connections", {}).items()
        ],
        "ports": [(p, c.split(",")[0]) for p, c in netlist.get("ports", {}).items()],
    }


def _get_connectivity_graph(netlist: sax.Netlist) -> nx.Graph:
    graph = nx.Graph()
    connectivity_netlist = _get_connectivity_netlist(netlist)
    for name in connectivity_netlist["instances"]:
        graph.add_node(name)
    for c1, c2 in connectivity_netlist["connections"]:
        graph.add_edge(c1, c2)
    for c1, c2 in connectivity_netlist["ports"]:
        graph.add_edge(c1, c2)
    return graph


def _flatten_netlist_into(  # noqa: PLR0912,C901
    recnet: sax.RecursiveNetlist, net: sax.Netlist, sep: str
) -> None:
    for name, instance in list(net["instances"].items()):
        component = instance["component"]
        if component not in recnet:
            continue
        del net["instances"][name]
        child_net = deepcopy(recnet[component])
        _flatten_netlist_into(recnet, child_net, sep)
        for iname, iinstance in child_net["instances"].items():
            net["instances"][f"{name}{sep}{iname}"] = iinstance
        ports = {k: f"{name}{sep}{v}" for k, v in child_net.get("ports", {}).items()}
        net["connections"] = net.get("connections", {})
        for ip1, ip2 in list(net["connections"].items()):
            n1, p1 = ip1.split(",")
            n2, p2 = ip2.split(",")
            if n1 == name:
                del net["connections"][ip1]
                if p1 not in ports:
                    warnings.warn(
                        f"Port {ip1} not found. Connection {ip1}<->{ip2} ignored.",
                        stacklevel=2,
                    )
                    continue
                net["connections"][ports[p1]] = ip2
            elif n2 == name:
                if p2 not in ports:
                    warnings.warn(
                        f"Port {ip2} not found. Connection {ip1}<->{ip2} ignored.",
                        stacklevel=2,
                    )
                    continue
                net["connections"][ip1] = ports[p2]
        child_net["connections"] = child_net.get("connections", {})
        for ip1, ip2 in child_net["connections"].items():
            net["connections"][f"{name}{sep}{ip1}"] = f"{name}{sep}{ip2}"
        for p, ip2 in list(net.get("ports", {}).items()):
            try:
                n2, p2 = ip2.split(",")
            except ValueError:
                warnings.warn(
                    f"Unconventional port definition ignored: {p}->{ip2}.", stacklevel=2
                )
                continue
            if n2 == name:
                if p2 in ports:
                    net.get("ports", {})[p] = ports[p2]
                else:
                    del net.get("ports", {})[p]


@overload
def convert_nets_to_connections(
    netlist: sax.Netlist,
) -> sax.Netlist: ...


@overload
def convert_nets_to_connections(
    netlist: sax.RecursiveNetlist,
) -> sax.RecursiveNetlist: ...


def convert_nets_to_connections(
    netlist: sax.AnyNetlist,
) -> sax.AnyNetlist:
    if (recnet := sax.try_into[sax.RecursiveNetlist](netlist)) is not None:
        return {k: convert_nets_to_connections(v) for k, v in recnet.items()}
    net: sax.Netlist = sax.into[sax.Netlist](netlist)
    nets = net.pop("nets", [])
    connections = net.get("connections", {})
    connections = _nets_to_connections(nets, connections)
    net["connections"] = connections
    return net


def _nets_to_connections(
    nets: sax.Nets, connections: sax.Connections
) -> sax.Connections:
    connections = dict(connections.items())
    inverse_connections = {v: k for k, v in connections.items()}

    def _is_connected(p: sax.InstancePort) -> bool:
        return (p in connections) or (p in inverse_connections)

    def _add_connection(p: sax.InstancePort, q: sax.InstancePort) -> None:
        connections[p] = q
        inverse_connections[q] = p

    def _get_connected_port(p: sax.InstancePort) -> sax.InstancePort:
        if p in connections:
            return connections[p]
        return inverse_connections[p]

    for net in nets:
        p = net["p1"]
        q = net["p2"]
        if _is_connected(p):
            _q = _get_connected_port(p)
            msg = (
                "SAX currently does not support multiply connected ports. "
                f"Got {p}<->{q} and {p}<->{_q}"
            )
            raise ValueError(
                msg,
            )
        if _is_connected(q):
            _p = _get_connected_port(q)
            msg = (
                "SAX currently does not support multiply connected ports. "
                f"Got {p}<->{q} and {_p}<->{q}"
            )
            raise ValueError(
                msg,
            )
        _add_connection(p, q)
    return connections


@overload
def expand_probes(
    netlist: sax.Netlist,
    probes: dict[str, str],
) -> sax.Netlist: ...


@overload
def expand_probes(
    netlist: sax.RecursiveNetlist,
    probes: dict[str, str],
) -> sax.RecursiveNetlist: ...


def expand_probes(
    netlist: sax.AnyNetlist,
    probes: dict[str, str],
) -> sax.AnyNetlist:
    """Expand probes into ideal_probe instances and update connections/ports.

    This function inserts ideal 4-port measurement probes at specified connection
    points in the netlist. Each probe intercepts a connection and exposes forward
    and backward traveling wave ports.

    Args:
        netlist: The netlist to expand probes in.
        probes: A mapping from probe names to instance ports where probes should
            be inserted. If the instance port is part of an existing connection,
            a 4-port probe is inserted. If the instance port is unconnected,
            only the "X_fwd" port is created as a direct alias.
            Example: ``{"mid": "wg1,out"}``

    Returns:
        A new netlist with probe instances inserted and connections/ports updated.
        For probes on connected ports, two new ports are added: "X_fwd" and "X_bwd".
        For probes on unconnected ports, only "X_fwd" is added.

    Raises:
        ValueError: If probe ports would conflict with existing ports.

    Example:
        ```python
        netlist = {
            "instances": {"wg1": "waveguide", "wg2": "waveguide"},
            "connections": {"wg1,out": "wg2,in"},
            "ports": {"in": "wg1,in", "out": "wg2,out"},
        }
        probes = {"mid": "wg1,out"}
        expanded = expand_probes(netlist, probes)
        # expanded now has instances: wg1, wg2, _probe_mid
        # expanded now has ports: in, out, mid_fwd, mid_bwd
        ```
    """
    if not probes:
        return netlist

    # Handle recursive netlist: only expand probes in top-level
    if (recnet := sax.try_into[sax.RecursiveNetlist](netlist)) is not None:
        top_level_name = next(iter(recnet))
        top_level = recnet[top_level_name]
        expanded_top = expand_probes(top_level, probes)
        return {
            top_level_name: expanded_top,
            **{k: v for k, v in recnet.items() if k != top_level_name},
        }

    # It's a flat netlist
    net: sax.Netlist = deepcopy(sax.into[sax.Netlist](netlist))
    connections = dict(net.get("connections", {}).items())
    inverse_connections = {v: k for k, v in connections.items()}
    ports = dict(net.get("ports", {}).items())
    instances = dict(net["instances"].items())

    for probe_name, instance_port in probes.items():
        # Validate probe name won't conflict with existing ports
        fwd_port = f"{probe_name}_fwd"
        bwd_port = f"{probe_name}_bwd"
        if fwd_port in ports or bwd_port in ports:
            msg = (
                f"Probe '{probe_name}' would create ports '{fwd_port}'/'{bwd_port}' "
                f"which conflict with existing ports."
            )
            raise ValueError(msg)

        # Validate instance name won't conflict
        probe_instance_name = f"_probe_{probe_name}"
        if probe_instance_name in instances:
            msg = (
                f"Probe instance name '{probe_instance_name}' conflicts with "
                f"an existing instance."
            )
            raise ValueError(msg)

        # Find the connection containing this instance port.
        # Track orientation so _fwd means "signal flowing INTO instance_port"
        if instance_port in connections:
            # instance_port is on the left side: instance_port -> other_port
            # Signal flows OUT of instance_port, so we need to flip the probe
            other_port = connections[instance_port]
            del connections[instance_port]
            del inverse_connections[other_port]
            # probe,in connects to other_port (source of signal INTO instance_port)
            # probe,out connects to instance_port (destination)
            in_side = other_port
            out_side = instance_port
        elif instance_port in inverse_connections:
            # instance_port is on the right side: other_port -> instance_port
            # Signal flows INTO instance_port, so probe is correctly oriented
            other_port = inverse_connections[instance_port]
            del connections[other_port]
            del inverse_connections[instance_port]
            # probe,in connects to other_port (source of signal INTO instance_port)
            # probe,out connects to instance_port (destination)
            in_side = other_port
            out_side = instance_port
        else:
            # Unconnected port: just expose it as a top-level port (fwd only)
            ports[fwd_port] = instance_port
            continue

        # Insert probe instance
        instances[probe_instance_name] = {"component": "_ideal_probe"}

        # Create new connections: in_side -> probe,in and probe,out -> out_side
        # This orients the probe so _fwd captures signal flowing INTO instance_port
        connections[in_side] = f"{probe_instance_name},in"
        inverse_connections[f"{probe_instance_name},in"] = in_side
        connections[f"{probe_instance_name},out"] = out_side
        inverse_connections[out_side] = f"{probe_instance_name},out"

        # Add probe ports
        ports[fwd_port] = f"{probe_instance_name},tap_fwd"
        ports[bwd_port] = f"{probe_instance_name},tap_bwd"

    net["instances"] = instances
    net["connections"] = connections
    net["ports"] = ports
    return net



"""SAX KLU Backend."""

from __future__ import annotations

from typing import Any

import jax
import jax.numpy as jnp
from natsort import natsorted

import sax

__all__ = [
    "analyze_circuit_klu",
    "analyze_instances_klu",
    "evaluate_circuit_klu",
]


def analyze_instances_klu(
    instances: dict[sax.InstanceName, sax.Instance],
    models: dict[str, sax.Model],
) -> dict[str, sax.SCoo]:
    """Analyze circuit instances for the KLU backend.

    Prepares instance S-matrices for the KLU backend by converting all component
    models to SCoo (coordinate) format. The KLU backend uses sparse matrix
    techniques with the KLU solver for high-performance circuit evaluation.

    Args:
        instances: Dictionary mapping instance names to instance definitions
            containing component names and settings.
        models: Dictionary mapping component names to their model functions.

    Returns:
        Dictionary mapping instance names to their S-matrices in SCoo format.

    Note:
        The KLU backend is the recommended high-performance backend for most
        circuit simulations. It uses sparse matrix factorization and can handle
        large circuits efficiently with full bidirectional coupling and reflections.

    Example:
        ```python
        instances = {
            "wg1": {"component": "waveguide", "settings": {"length": 10.0}},
            "dc1": {"component": "coupler", "settings": {"coupling": 0.1}},
        }
        models = {"waveguide": waveguide_model, "coupler": coupler_model}
        analyzed = analyze_instances_klu(instances, models)
        ```
    """
    instances = sax.into[sax.Instances](instances)
    model_names = set()
    for i in instances.values():
        model_names.add(i["component"])
    dummy_models = {k: sax.scoo(models[k]()) for k in model_names}
    dummy_instances = {}
    for k, i in instances.items():
        dummy_instances[k] = dummy_models[i["component"]]
    return dummy_instances


def analyze_circuit_klu(
    analyzed_instances: dict[sax.InstanceName, sax.SCoo],
    connections: sax.Connections,
    ports: sax.Ports,
) -> Any:  # noqa: ANN401
    """Analyze circuit topology for the KLU sparse matrix backend.

    Performs detailed circuit analysis to set up the sparse matrix system for
    the KLU solver. This analysis creates the connection matrices and index
    mappings needed for efficient sparse matrix operations.

    Args:
        analyzed_instances: Instance S-matrices from analyze_instances_klu in
            SCoo format.
        connections: Dictionary mapping instance ports to each other, defining
            internal circuit connections.
        ports: Dictionary mapping external port names to instance ports.

    Returns:
        Complex analysis data structure containing sparse matrix indices,
        connection mappings, and external port information optimized for
        the KLU solver.

    Note:
        This analysis step is computationally intensive but enables very fast
        circuit evaluation, especially for large circuits with many components.

    Example:
        ```python
        connections = {"wg1,out": "dc1,in1", "dc1,out1": "wg2,in"}
        ports = {"in": "wg1,in", "out": "wg2,out"}
        analyzed = analyze_circuit_klu(analyzed_instances, connections, ports)
        ```
    """
    connections = {**connections, **{v: k for k, v in connections.items()}}
    inverse_ports = {v: k for k, v in ports.items()}
    port_map = {k: i for i, k in enumerate(ports)}

    idx, Si, Sj, instance_ports = 0, [], [], {}
    for name, instance in analyzed_instances.items():
        si, sj, _, ports_map = sax.scoo(instance)
        Si.append(si + idx)
        Sj.append(sj + idx)
        instance_ports.update({f"{name},{p}": i + idx for p, i in ports_map.items()})
        idx += len(ports_map)

    n_col = idx
    n_rhs = len(port_map)

    Si = jnp.concatenate(Si, -1)
    Sj = jnp.concatenate(Sj, -1)

    Cmap = {
        int(instance_ports[k]): int(instance_ports[v]) for k, v in connections.items()
    }
    Ci = jnp.array(list(Cmap.keys()), dtype=jnp.int32)
    Cj = jnp.array(list(Cmap.values()), dtype=jnp.int32)

    Cextmap = {
        int(instance_ports[k]): int(port_map[v]) for k, v in inverse_ports.items()
    }
    Cexti = jnp.stack(list(Cextmap.keys()), 0)
    Cextj = jnp.stack(list(Cextmap.values()), 0)
    Cext = jnp.zeros((n_col, n_rhs), dtype=complex).at[Cexti, Cextj].set(1.0)

    mask = Cj[None, :] == Si[:, None]
    CSi = jnp.broadcast_to(Ci[None, :], mask.shape)[mask]

    mask = (Cj[:, None] == Si[None, :]).any(0)
    CSj = Sj[mask]

    Ii = Ij = jnp.arange(n_col)
    I_CSi = jnp.concatenate([CSi, Ii], -1)
    I_CSj = jnp.concatenate([CSj, Ij], -1)
    return (
        n_col,
        mask,
        Si,
        Sj,
        Cext,
        Cexti,
        Cextj,
        I_CSi,
        I_CSj,
        tuple((k, v[1]) for k, v in analyzed_instances.items()),
        tuple(port_map),
    )


def evaluate_circuit_klu(
    analyzed: Any,  # noqa: ANN401
    instances: dict[sax.InstanceName, sax.SType],
) -> sax.SDense:
    """Evaluate circuit S-matrix using the KLU sparse matrix solver.

    Computes the circuit S-matrix by solving the sparse linear system
    (I - CS)x = C_ext using the high-performance KLU sparse matrix solver.
    This approach is highly efficient for large circuits.

    The algorithm:
    1. Assembles the sparse connection matrix C and S-matrix blocks
    2. Forms the system matrix (I - CS) where I is identity
    3. Solves the linear system using KLU factorization
    4. Extracts the external port S-matrix from the solution

    Args:
        analyzed: Complex analysis data from analyze_circuit_klu containing
            pre-computed sparse matrix indices and mappings.
        instances: Dictionary mapping instance names to their evaluated S-matrices
            in any SAX format (will be converted to SCoo).

    Returns:
        Circuit S-matrix in SDense format (dense matrix with port mapping).

    Note:
        This backend provides the best performance for most circuits, especially
        large ones. It handles all types of coupling (forward, backward, cross)
        and reflections accurately using sparse matrix techniques.

    Example:
        ```python
        # Circuit analysis and instances
        instances = {
            "wg1": {("in", "out"): 0.95 * jnp.exp(1j * 0.1)},
            "dc1": {
                ("in1", "out1"): 0.9,
                ("in1", "out2"): 0.1,
                ("in2", "out1"): 0.1,
                ("in2", "out2"): 0.9,
            },
        }
        circuit_s_matrix, port_map = evaluate_circuit_klu(analyzed, instances)
        # Result is a dense S-matrix with full coupling terms
        ```
    """
    import klujax

    (
        n_col,
        mask,
        Si,
        Sj,
        Cext,
        Cexti,
        Cextj,
        I_CSi,
        I_CSj,
        dummy_pms,
        port_map,
    ) = analyzed

    idx = 0
    Sx = []
    batch_shape = ()
    for name, _ in dummy_pms:
        _, _, sx, ports_map = sax.scoo(instances[name])
        Sx.append(sx)
        if len(sx.shape[:-1]) > len(batch_shape):
            batch_shape = sx.shape[:-1]
        idx += len(ports_map)

    Sx = jnp.concatenate(
        [jnp.broadcast_to(sx, (*batch_shape, sx.shape[-1])) for sx in Sx], -1
    )
    CSx = Sx[..., mask]
    Ix = jnp.ones((*batch_shape, n_col))
    I_CSx = jnp.concatenate([-CSx, Ix], -1)

    Sx = Sx.reshape(-1, Sx.shape[-1])  # n_lhs x N
    I_CSx = I_CSx.reshape(-1, I_CSx.shape[-1])  # n_lhs x M
    solve_klu = jax.vmap(klujax.solve, (None, None, 0, None), 0)
    inv_I_CS_Cext = solve_klu(I_CSi, I_CSj, I_CSx, Cext)
    mul_coo = jax.vmap(klujax.dot, (None, None, 0, 0), 0)
    S_inv_I_CS_Cext = mul_coo(Si, Sj, Sx, inv_I_CS_Cext)

    CextT_S_inv_I_CS_Cext = S_inv_I_CS_Cext[..., Cexti, :][..., :, Cextj]

    _, n, _ = CextT_S_inv_I_CS_Cext.shape
    S = CextT_S_inv_I_CS_Cext.reshape(*batch_shape, n, n)

    return S, {p: i for i, p in enumerate(port_map)}


def _get_instance_ports(
    connections: dict[str, str], ports: dict[str, str]
) -> dict[str, list[str]]:
    instance_ports = {}
    for connection in connections.items():
        for ip in connection:
            i, p = ip.split(",")
            if i not in instance_ports:
                instance_ports[i] = set()
            instance_ports[i].add(p)
    for ip in ports.values():
        i, p = ip.split(",")
        if i not in instance_ports:
            instance_ports[i] = set()
        instance_ports[i].add(p)
    return {k: natsorted(v) for k, v in instance_ports.items()}


def _get_dummy_instances(
    connections: dict[str, str],
    ports: dict[str, str],
) -> dict[str, tuple[Any, Any, None, dict[str, int]]]:
    """No longer used. deprecated by analyze_instances_klu."""
    instance_ports = _get_instance_ports(connections, ports)
    dummy_instances = {}
    for name, _ports in instance_ports.items():
        num_ports = len(_ports)
        pm = {p: i for i, p in enumerate(_ports)}
        ij = jnp.mgrid[:num_ports, :num_ports]
        i = ij[0].ravel()
        j = ij[1].ravel()
        dummy_instances[name] = (i, j, None, pm)
    return dummy_instances



"""SAX MMI Models."""

import jax
import jax.numpy as jnp
from pydantic import validate_call

import sax

from .couplers import coupler_ideal
from .splitters import splitter_ideal


@jax.jit
@validate_call
def mmi1x2_ideal(wl: sax.FloatArrayLike = sax.WL_C) -> sax.SDict:
    """Ideal 1x2 multimode interference (MMI) splitter model.

    ```{svgbob}
            +-------------+     out1
            |             |---* o2
     o1 *---|             |
    in0     |             |---* o3
            +-------------+     out0
    ```

    Args:
        wl: The wavelength in micrometers.

    Returns:
        S-matrix dictionary representing the ideal MMI splitter behavior.

    Examples:
        Ideal 1x2 MMI:

        ```python
        # mkdocs: render
        import matplotlib.pyplot as plt
        import numpy as np
        import sax

        sax.set_port_naming_strategy("optical")

        wl = sax.wl_c()
        s = sax.models.mmi1x2_ideal(wl=wl)
        thru = np.abs(s[("o1", "o3")]) ** 2
        cross = np.abs(s[("o1", "o2")]) ** 2
        plt.figure()
        plt.plot(wl, thru, label="thru")
        plt.plot(wl, cross, label="cross")
        plt.xlabel("Wavelength [μm]")
        plt.ylabel("Power")
        plt.legend()
        ```
    """
    return splitter_ideal(wl=wl, coupling=0.5)


@jax.jit
@validate_call
def mmi2x2_ideal(*, wl: sax.FloatArrayLike = sax.WL_C) -> sax.SDict:
    """Ideal 2x2 multimode interference (MMI) coupler model.

    ```{svgbob}
    in1     +-------------+     out1
     o2 *---|             |---* o3
            |             |
     o1 *---|             |---* o4
    in0     +-------------+     out0
    ```

    Args:
        wl: The wavelength in micrometers.

    Returns:
        S-matrix dictionary representing the ideal MMI coupler behavior.

    Examples:
        Ideal 2x2 MMI:

        ```python
        # mkdocs: render
        import matplotlib.pyplot as plt
        import numpy as np
        import sax

        sax.set_port_naming_strategy("optical")

        wl = sax.wl_c()
        s = sax.models.mmi2x2_ideal(wl=wl)
        thru = np.abs(s[("o1", "o4")]) ** 2
        cross = np.abs(s[("o1", "o3")]) ** 2
        plt.figure()
        plt.plot(wl, thru, label="thru")
        plt.plot(wl, cross, label="cross")
        plt.xlabel("Wavelength [μm]")
        plt.ylabel("Power")
        plt.legend()
        ```
    """
    return coupler_ideal(wl=wl, coupling=0.5)


@jax.jit
@validate_call
def mmi1x2(
    wl: sax.FloatArrayLike = sax.WL_C,
    wl0: sax.FloatArrayLike = sax.WL_C,
    fwhm: sax.FloatArrayLike = 0.2,
    loss_dB: sax.FloatArrayLike = 0.3,
) -> sax.SDict:
    r"""Realistic 1x2 MMI splitter model with dispersion and loss.

    ```{svgbob}
            +-------------+     out1
            |             |---* o2
     o1 *---|             |
    in0     |             |---* o3
            +-------------+     out0
    ```

    Args:
        wl: The wavelength in micrometers.
        wl0: The Center wavelength of the MMI
        fwhm: The Full width at half maximum or the MMI.
        loss_dB: Insertion loss in dB at the center wavelength.

    Returns:
        S-matrix dictionary representing the dispersive MMI splitter behavior.

    Examples:
        Basic 1x2 MMI:

        ```python
        # mkdocs: render
        import matplotlib.pyplot as plt
        import numpy as np
        import sax

        sax.set_port_naming_strategy("optical")

        wl = sax.wl_c()
        s = sax.models.mmi1x2(wl=wl)
        thru = np.abs(s[("o1", "o3")]) ** 2
        cross = np.abs(s[("o1", "o2")]) ** 2
        plt.figure()
        plt.plot(wl, thru, label="thru")
        plt.plot(wl, cross, label="cross")
        plt.xlabel("Wavelength [μm]")
        plt.ylabel("Power")
        plt.legend()
        ```
    """
    thru = _mmi_amp(wl=wl, wl0=wl0, fwhm=fwhm, loss_dB=loss_dB) / 2**0.5

    p = sax.PortNamer(1, 2)
    return sax.reciprocal(
        {
            (p.o1, p.o2): thru,
            (p.o1, p.o3): thru,
        }
    )


@jax.jit
@validate_call
def mmi2x2(
    wl: sax.FloatArrayLike = sax.WL_C,
    wl0: sax.FloatArrayLike = sax.WL_C,
    fwhm: sax.FloatArrayLike = 0.2,
    loss_dB: sax.FloatArrayLike = 0.3,
    shift: sax.FloatArrayLike = 0.0,
    loss_dB_cross: sax.FloatArrayLike | None = None,
    loss_dB_thru: sax.FloatArrayLike | None = None,
    splitting_ratio_cross: sax.FloatArrayLike = 0.5,
    splitting_ratio_thru: sax.FloatArrayLike = 0.5,
) -> sax.SDict:
    r"""Realistic 2x2 MMI coupler model with dispersion and asymmetry.

    ```{svgbob}
    in1     +-------------+     out1
     o2 *---|             |---* o3
            |             |
     o1 *---|             |---* o4
    in0     +-------------+     out0
    ```

    Args:
        wl: wavelength in micrometers.
        wl0: Center wavelength of the MMI in micrometers.
        fwhm: Full width at half maximum bandwidth in micrometers.
        loss_dB: Insertion loss of the MMI at center wavelength.
        shift: The peak shift of the cross-transmission in micrometers.
        loss_dB_cross: Optional separate insertion loss in dB for cross ports.
            If None, uses loss_dB. Allows modeling of asymmetric loss.
        loss_dB_thru: Optional separate insertion loss in dB for bar (through)
            ports. If None, uses loss_dB. Allows modeling of asymmetric loss.
        splitting_ratio_cross: Power splitting ratio for cross ports (0 to 1).
            Allows modeling of imbalanced coupling. Defaults to 0.5.
        splitting_ratio_thru: Power splitting ratio for bar ports (0 to 1).
            Allows modeling of imbalanced transmission. Defaults to 0.5.

    Returns:
        S-matrix dictionary representing the realistic MMI coupler behavior.

    Examples:
        Basic 2x2 MMI:

        ```python
        # mkdocs: render
        import matplotlib.pyplot as plt
        import numpy as np
        import sax

        sax.set_port_naming_strategy("optical")

        wl = sax.wl_c()
        s = sax.models.mmi2x2(wl=wl, shift=0.001)
        thru = np.abs(s[("o1", "o4")]) ** 2
        cross = np.abs(s[("o1", "o3")]) ** 2
        plt.figure()
        plt.plot(wl, thru, label="thru")
        plt.plot(wl, cross, label="cross")
        plt.xlabel("Wavelength [μm]")
        plt.ylabel("Power")
        plt.legend()
        ```
    ```
    """
    loss_dB_cross = loss_dB_cross or loss_dB
    loss_dB_thru = loss_dB_thru or loss_dB

    # Convert splitting ratios from power to amplitude by taking the square root
    amplitude_ratio_thru = splitting_ratio_thru**0.5
    amplitude_ratio_cross = splitting_ratio_cross**0.5

    # _mmi_amp already includes the loss, so we don't need to apply it again
    thru = (
        _mmi_amp(wl=wl, wl0=wl0, fwhm=fwhm, loss_dB=loss_dB_thru) * amplitude_ratio_thru
    )
    cross = (
        1j
        * _mmi_amp(wl=wl, wl0=wl0 + shift, fwhm=fwhm, loss_dB=loss_dB_cross)
        * amplitude_ratio_cross
    )

    p = sax.PortNamer(2, 2)
    return sax.reciprocal(
        {
            (p.o1, p.o3): thru,
            (p.o1, p.o4): cross,
            (p.o2, p.o3): cross,
            (p.o2, p.o4): thru,
        }
    )


def _mmi_amp(
    wl: sax.FloatArrayLike = sax.WL_C,
    wl0: sax.FloatArrayLike = sax.WL_C,
    fwhm: sax.FloatArrayLike = 0.2,
    loss_dB: sax.FloatArrayLike = 0.3,
) -> sax.FloatArray:
    max_amplitude = 10 ** (-abs(loss_dB) / 20)
    f = 1 / wl
    f0 = 1 / wl0
    f1 = 1 / (wl0 + fwhm / 2)
    f2 = 1 / (wl0 - fwhm / 2)
    _fwhm = f2 - f1

    sigma = _fwhm / (2 * jnp.sqrt(2 * jnp.log(2)))
    # Gaussian response in frequency domain
    spectral_response = jnp.exp(-((f - f0) ** 2) / (2 * sigma**2))
    # Apply loss to amplitude, not power
    amplitude = max_amplitude * spectral_response / spectral_response.max()
    return jnp.asarray(amplitude)


def _mmi_nxn(
    n: int,
    wl: sax.FloatArrayLike = sax.WL_C,
    wl0: sax.FloatArrayLike = sax.WL_C,
    fwhm: sax.FloatArrayLike = 0.2,
    loss_dB: sax.FloatArrayLike | None = None,
    shift: sax.FloatArrayLike | None = None,
    splitting_matrix: sax.FloatArray2D | None = None,
) -> sax.SDict:
    _loss_dB = jnp.zeros(n) if loss_dB is None else jnp.asarray(loss_dB)
    _shift = jnp.zeros(n) if shift is None else jnp.asarray(shift)
    _splitting_matrix = (
        jnp.full((n, n), 1 / n)
        if splitting_matrix is None
        else jnp.asarray(splitting_matrix)
    )

    S = {}
    p = sax.PortNamer(n, n)
    for i in range(n):
        for j in range(n):
            amplitude = _mmi_amp(wl, wl0 + _shift[j], fwhm, _loss_dB[j])
            amplitude *= jnp.sqrt(_splitting_matrix[i][j])
            # _mmi_amp already includes the loss, so no additional loss factor needed
            S[(p[i], p[n + j])] = amplitude

    return sax.reciprocal(S)
